from __future__ import annotations

import enum
import functools
import typing as t

import attrs
import mitsuba

from .attrs import documented, frozen
from .exceptions import UnsetModeError, UnsupportedModeError

# ------------------------------------------------------------------------------
#                                 Mode flags
# ------------------------------------------------------------------------------


class ModeFlag(enum.Flag):
    """
    Flags defining the various possible features associated to an Eradiate mode.
    """

    NONE = 0  #: No flags
    SPECTRAL_MODE_MONO = enum.auto()  #: Monochromatic (line-by-line) mode
    SPECTRAL_MODE_CKD = enum.auto()  #: Correlated-k distribution mode
    MI_BACKEND_SCALAR = enum.auto()  #: Mitsuba scalar backend
    MI_BACKEND_LLVM = enum.auto()  #: Mitsuba LLVM backend
    MI_COLOR_MODE_MONO = enum.auto()  #: Mitsuba monochromatic mode
    MI_DOUBLE_PRECISION_NO = enum.auto()  #: Mitsuba single-precision
    MI_DOUBLE_PRECISION_YES = enum.auto()  #: Mitsuba double-precision
    MI_POLARIZED_NO = enum.auto()  #: Unpolarized
    MI_POLARIZED_YES = enum.auto()  #: Polarized
    ANY = (
        SPECTRAL_MODE_MONO
        | SPECTRAL_MODE_CKD
        | MI_BACKEND_SCALAR
        | MI_BACKEND_LLVM
        | MI_COLOR_MODE_MONO
        | MI_DOUBLE_PRECISION_NO
        | MI_DOUBLE_PRECISION_YES
        | MI_POLARIZED_NO
        | MI_POLARIZED_YES
    )  #: All flags


# ------------------------------------------------------------------------------
#                              Mode definitions
# ------------------------------------------------------------------------------


# Map associating a mode ID string to the corresponding class
# (aliased in public API section). This is implemented as a cached function to
# make a delayed evaluation possible (otherwise, the Mode class is not defined).
# See also the Mode.new() constructor.
@functools.lru_cache(maxsize=1)
def _mode_registry() -> dict[str, Mode]:
    return {
        k: Mode(id=k, **v)
        for k, v in {
            "mono_single": {
                "spectral_mode": "mono",
                "mi_backend": "scalar",
                "mi_color_mode": "mono",
                "mi_double_precision": False,
                "mi_polarized": False,
            },
            "mono_polarized_single": {
                "spectral_mode": "mono",
                "mi_backend": "scalar",
                "mi_color_mode": "mono",
                "mi_double_precision": False,
                "mi_polarized": True,
            },
            "mono_double": {
                "spectral_mode": "mono",
                "mi_backend": "scalar",
                "mi_color_mode": "mono",
                "mi_double_precision": True,
                "mi_polarized": False,
            },
            "mono_polarized_double": {
                "spectral_mode": "mono",
                "mi_backend": "scalar",
                "mi_color_mode": "mono",
                "mi_double_precision": True,
                "mi_polarized": True,
            },
            "ckd_single": {
                "spectral_mode": "ckd",
                "mi_backend": "scalar",
                "mi_color_mode": "mono",
                "mi_double_precision": False,
                "mi_polarized": False,
            },
            "ckd_polarized_single": {
                "spectral_mode": "ckd",
                "mi_backend": "scalar",
                "mi_color_mode": "mono",
                "mi_double_precision": False,
                "mi_polarized": True,
            },
            "ckd_double": {
                "spectral_mode": "ckd",
                "mi_backend": "scalar",
                "mi_color_mode": "mono",
                "mi_double_precision": True,
                "mi_polarized": False,
            },
            "ckd_polarized_double": {
                "spectral_mode": "ckd",
                "mi_backend": "scalar",
                "mi_color_mode": "mono",
                "mi_double_precision": True,
                "mi_polarized": True,
            },
        }.items()
    }


def _spectral_mode_converter(value: str | ModeFlag):
    if isinstance(value, str):
        value = value.upper()
        if not value.startswith("SPECTRAL_MODE_"):
            value = f"SPECTRAL_MODE_{value}"
        return ModeFlag[value]
    else:
        return ModeFlag(value)


def _mi_backend_converter(value: str | ModeFlag):
    if isinstance(value, str):
        value = value.upper()
        if not value.startswith("MI_BACKEND_"):
            value = f"MI_BACKEND_{value}"
        return ModeFlag[value]
    else:
        return ModeFlag(value)


def _mi_color_mode_converter(value: str | ModeFlag):
    if isinstance(value, str):
        value = value.upper()
        if not value.startswith("MI_COLOR_MODE_"):
            value = f"MI_COLOR_MODE_{value}"
        return ModeFlag[value]
    else:
        return ModeFlag(value)


def _mi_polarized_converter(value: bool | ModeFlag):
    if isinstance(value, bool):
        return ModeFlag.MI_POLARIZED_YES if value is True else ModeFlag.MI_POLARIZED_NO
    else:
        return ModeFlag(value)


def _mi_double_precision_converter(value: bool | ModeFlag):
    if isinstance(value, bool):
        return (
            ModeFlag.MI_DOUBLE_PRECISION_YES
            if value is True
            else ModeFlag.MI_DOUBLE_PRECISION_NO
        )
    else:
        return ModeFlag(value)


@frozen
class Mode:
    """
    Data structure describing Eradiate's operational mode and associated
    ancillary data.

    Warnings
    --------
    Instances are immutable.
    """

    id: str = documented(
        attrs.field(converter=attrs.converters.optional(str)),
        doc="Mode identifier.",
        type="str",
    )

    spectral_mode: t.Literal[
        ModeFlag.SPECTRAL_MODE_MONO, ModeFlag.SPECTRAL_MODE_CKD
    ] = documented(
        attrs.field(
            converter=_spectral_mode_converter,
            validator=attrs.validators.in_(
                {ModeFlag.SPECTRAL_MODE_MONO, ModeFlag.SPECTRAL_MODE_CKD}
            ),
        ),
        doc="Spectral dimension handling.",
        type=".ModeFlag",
        init_type=".ModeFlag or str",
    )

    mi_backend: t.Literal[ModeFlag.MI_BACKEND_SCALAR, ModeFlag.MI_BACKEND_LLVM] = (
        documented(
            attrs.field(
                converter=_mi_backend_converter,
                validator=attrs.validators.in_(
                    {ModeFlag.MI_BACKEND_SCALAR, ModeFlag.MI_BACKEND_LLVM}
                ),
            ),
            doc="Mitsuba computational backend.",
            type=".ModeFlag",
            init_type=".ModeFlag or str",
        )
    )

    mi_color_mode: t.Literal[ModeFlag.MI_COLOR_MODE_MONO] = documented(
        attrs.field(
            converter=_mi_color_mode_converter,
            validator=attrs.validators.in_({ModeFlag.MI_COLOR_MODE_MONO}),
        ),
        doc="Mitsuba color mode.",
        type=".ModeFlag",
        init_type=".ModeFlag or str",
    )

    mi_polarized: t.Literal[ModeFlag.MI_POLARIZED_NO, ModeFlag.MI_POLARIZED_YES] = (
        documented(
            attrs.field(
                converter=_mi_polarized_converter,
                validator=attrs.validators.in_(
                    {ModeFlag.MI_POLARIZED_NO, ModeFlag.MI_POLARIZED_YES}
                ),
            ),
            doc="Mitsuba polarized mode.",
            type=".ModeFlag",
            init_type=".ModeFlag or str",
        )
    )

    mi_double_precision: t.Literal[
        ModeFlag.MI_DOUBLE_PRECISION_NO, ModeFlag.MI_DOUBLE_PRECISION_YES
    ] = documented(
        attrs.field(
            converter=_mi_double_precision_converter,
            validator=attrs.validators.in_(
                {ModeFlag.MI_DOUBLE_PRECISION_NO, ModeFlag.MI_DOUBLE_PRECISION_YES}
            ),
        ),
        doc="Mitsuba double precision.",
        type=".ModeFlag",
        init_type=".ModeFlag or str",
    )

    @property
    def mi_flags(self) -> ModeFlag:
        """
        Combined Mitsuba-specific flags.
        """
        return (
            self.mi_backend
            | self.mi_color_mode
            | self.mi_polarized
            | self.mi_double_precision
        )

    @property
    def flags(self) -> ModeFlag:
        """
        All flags combined.
        """
        return self.spectral_mode | self.mi_flags

    @property
    def mi_variant(self) -> str:
        """
        Mitsuba variant associated with the selected mode.
        """
        result = [
            self.mi_backend.name.lower().removeprefix("mi_backend_"),
            self.mi_color_mode.name.lower().removeprefix("mi_color_mode_"),
        ]
        if self.mi_polarized is ModeFlag.MI_POLARIZED_YES:
            result.append("polarized")
        if self.mi_double_precision is ModeFlag.MI_DOUBLE_PRECISION_YES:
            result.append("double")
        return "_".join(result)

    def check(
        self,
        spectral_mode: ModeFlag | str = ModeFlag.NONE,
        mi_backend: ModeFlag | str = ModeFlag.NONE,
        mi_color_mode: ModeFlag | str = ModeFlag.NONE,
        mi_polarized: ModeFlag | bool = ModeFlag.NONE,
        mi_double_precision: ModeFlag | bool = ModeFlag.NONE,
    ) -> ModeFlag:
        """
        Check if the currently active mode has the passed flags.

        Parameters
        ----------
        spectral_mode : .ModeFlag or str, optional
            Spectral mode to check. If unset, the check is skipped.

        mi_backend : .ModeFlag or str, optional
            Mitsuba backend to check. If unset, the check is skipped.

        mi_color_mode : .ModeFlag or str, optional
            Mitsuba color mode to check. If unset, the check is skipped.

        mi_polarized : .ModeFlag or bool, optional
            Mitsuba polarized mode to check. If unset, the check is skipped.

        mi_double_precision : .ModeFlag or bool, optional
            Mitsuba double precision mode to check. If unset, the check is skipped.

        Returns
        -------
        bool
            ``True`` if current mode has the passed flags, ``False`` otherwise.
        """
        condition = (
            _spectral_mode_converter(spectral_mode)
            | _mi_backend_converter(mi_backend)
            | _mi_color_mode_converter(mi_color_mode)
            | _mi_polarized_converter(mi_polarized)
            | _mi_double_precision_converter(mi_double_precision)
        )
        return condition & self.flags

    @property
    def is_mono(self) -> bool:
        return self.spectral_mode is ModeFlag.SPECTRAL_MODE_MONO

    @property
    def is_ckd(self) -> bool:
        return self.spectral_mode is ModeFlag.SPECTRAL_MODE_CKD

    @property
    def is_single_precision(self) -> bool:
        return self.mi_double_precision is ModeFlag.MI_DOUBLE_PRECISION_NO

    @property
    def is_double_precision(self) -> bool:
        return self.mi_double_precision is ModeFlag.MI_DOUBLE_PRECISION_YES

    @property
    def is_polarized(self) -> bool:
        return self.mi_polarized is ModeFlag.MI_POLARIZED_YES

    @staticmethod
    def new(mode_id: str) -> Mode:
        """
        Create a :class:`Mode` instance given its identifier. Available modes are:

        * ``mono``: Alias to ``mono_double``
        * ``ckd``: Alias to ``ckd_double``
        * ``mono_polarized``: Alias to ``mono_polarized_double``
        * ``ckd_polarized``: Alias to ``ckd_polarized_double``
        * ``mono_single``: Monochromatic, single-precision
        * ``mono_polarized_single``: Monochromatic, polarized, single-precision
        * ``mono_double``: Monochromatic, double-precision
        * ``mono_polarized_double``: Monochromatic, polarized, double-precision
        * ``ckd_single``: CKD, single-precision
        * ``ckd_polarized_single``: CKD, polarized, single-precision
        * ``ckd_double``: CKD, double-precision
        * ``ckd_polarized_double``: CKD, polarized, double-precision

        Parameters
        ----------
        mode_id : str
            String identifier for the created :class:`Mode` instance.

        Returns
        -------
        .Mode
            Created :class:`.Mode` instance.
        """
        try:
            return _mode_registry()[mode_id]
        except KeyError:
            raise ValueError(f"unknown mode '{mode_id}'")


# Define mode aliases
_mode_registry().update(
    {
        "mono": _mode_registry()["mono_double"],
        "mono_polarized": _mode_registry()["mono_polarized_double"],
        "ckd": _mode_registry()["ckd_double"],
        "ckd_polarized": _mode_registry()["ckd_polarized_double"],
    }
)

# Eradiate's operational mode configuration
_active_mode: Mode | None = None


# ------------------------------------------------------------------------------
#                            Mode subtype dispatcher
# ------------------------------------------------------------------------------


@attrs.define
class SubtypeDispatcher:
    """
    This is a very simple factory intended to map mode-specific flags to
    mode-dependent subtypes.
    Types can be registered with the :meth:`.register` method, and the
    appropriate subtype can be resolved with the :meth:`.resolve` method.
    """

    _type_name: str = attrs.field()
    _registry: dict[ModeFlag, t.Type] = attrs.field(factory=dict)

    def register(self, mode_flags: ModeFlag | str) -> t.Callable[[t.Type], t.Type]:
        """
        Register a subtype against a combination of mode flags. This method is
        meant to be used as a decorator.

        Parameters
        ----------
        mode_flags : .ModeFlag or str
            Mode flags against which the subtype is registered. If a string is
            passed, it is converted to a :class:`.ModeFlag`.
        """
        if isinstance(mode_flags, str):
            mode_flags = ModeFlag[mode_flags.upper()]

        def wrapper(cls):
            self._registry[mode_flags] = cls
            return cls

        return wrapper

    def resolve(self, mode_flags: ModeFlag | None = None) -> t.Type:
        """
        Resolve the subtype based against a set of mode flags.

        Parameters
        ----------
        mode_flags : .ModeFlag, optional
            A mode flag combination used to search the dispatcher's registry.
            The first entry that validates the flag conditions is returned,
            meaning that conflicting or redundant conditions will cause issues.
            If unspecified, the flags of the currently active mode are used.

        See Also
        --------
        :attr:`.Mode.flags`, :class:`.ModeFlag`
        """
        active_mode = get_mode()

        if mode_flags is None:
            mode_flags = active_mode.flags

        for key, value in self._registry.items():
            if mode_flags & key:
                return value

        raise NotImplementedError(
            f"Type {self._type_name} has no registered subtype for mode flags "
            f"{mode_flags}."
        )


# ------------------------------------------------------------------------------
#                                Public API
# ------------------------------------------------------------------------------


def get_mode(raise_exc: bool = True) -> Mode | None:
    """
    Get current operational mode.

    Parameters
    ----------
    raise_exc : bool, default: True
        If ``True``, raise an exception if mode has not been set prior
        (otherwise, this function will return ``None`` in such cases).

    Returns
    -------
    .Mode or None
        Current operational mode.

    Raises
    ------
    UnsetModeError
        If no active mode has been selected.

    See Also
    --------
    :func:`set_mode`
    """
    if _active_mode is None and raise_exc is True:
        raise UnsetModeError
    return _active_mode


def mode(raise_exc: bool = True) -> Mode | None:
    """This is a thin compatibility wrapper around :func:`.get_mode()`."""
    return get_mode(raise_exc=raise_exc)


def modes(
    filter: t.Callable[[Mode], bool] | None = None, asdict: bool = False
) -> list[str] | dict[str, Mode]:
    """
    Get list of registered operational modes.

    Parameters
    ----------
    filter : callable, optional
        A callable used to filter the returned modes. Operates on a
        :class:`.Mode` instance.

    asdict : bool, default: False
        If ``True``, returns a mode ID → mode instance dictionary; otherwise,
        returns a list of mode IDs only.

    Returns
    -------
    modes : list[str] or dict[str, .Mode]
        List of registered operational modes.

    Examples
    --------
    Return the full list of registered modes:

    >>> eradiate.modes()

    Return only CKD modes:

    >>> eradiate.modes(lambda x: x.is_ckd)
    """
    if filter is None:
        filter = lambda x: True  # noqa: E731

    if not asdict:
        return [k for k, v in _mode_registry().items() if filter(v)]
    else:
        return {k: v for k, v in _mode_registry().items() if filter(v)}


def set_mode(mode_id: str):
    """
    Set Eradiate's operational mode.

    This function sets and configures Eradiate's operational mode. Eradiate's
    modes map to Mitsuba's variants and are used to make contextual decisions
    when relevant during the translation of a scene to its kernel format.

    .. admonition:: Valid mode IDs
       :class: info

       * ``mono``: Alias to ``mono_double``
       * ``ckd``: Alias to ``ckd_double``
       * ``mono_polarized``: Alias to ``mono_polarized_double``
       * ``ckd_polarized``: Alias to ``ckd_polarized_double``
       * ``mono_single``: Monochromatic, single-precision
       * ``mono_polarized_single``: Monochromatic, polarized, single-precision
       * ``mono_double``: Monochromatic, double-precision
       * ``mono_polarized_double``: Monochromatic, polarized, double-precision
       * ``ckd_single``: CKD, single-precision
       * ``ckd_polarized_single``: CKD, polarized, single-precision
       * ``ckd_double``: CKD, double-precision
       * ``ckd_polarized_double``: CKD, polarized, double-precision
       * ``none`` (no mode selected)

    Parameters
    ----------
    mode_id : str
        Mode to be selected (see list below).

    Raises
    ------
    ValueError
        ``mode_id`` does not match any of the known mode identifiers.
    """
    global _active_mode

    if mode_id in _mode_registry():
        mode = Mode.new(mode_id)
        mitsuba.set_variant(mode.mi_variant)
    elif mode_id.lower() == "none":
        mode = None
    else:
        raise ValueError(f"unknown mode '{mode_id}'")

    _active_mode = mode


def supported_mode(**kwargs):
    """
    Check whether the current mode has specific features. If not, raise.

    Parameters
    ----------
    kwargs
        Keyword arguments passed to :meth:`.Mode.check`.

    Raises
    ------
    UnsupportedModeError
        Current mode does not pass the check.
    """
    if mode() is None or not mode().check(**kwargs):
        raise UnsupportedModeError


def unsupported_mode(**kwargs):
    """
    Check whether the current mode has specific features. If so, raise.

    Parameters
    ----------
    kwargs
        Keyword arguments passed to :meth:`.Mode.check`.

    Raises
    ------
    UnsupportedModeError
        Current mode has the requested flags.
    """
    if mode() is None or mode().check(**kwargs):
        raise UnsupportedModeError



"""
Spectral index data structure.

A spectral index data structure holds the information required to evaluate
spectral quantities. The simplest example is provided by
:class:`MonoSpectralIndex` which holds a single wavelength value.
Each SpectralIndex subclass is specific to an Eradiate spectral mode.
For example, :class:`MonoSpectralIndex` is used in monochromatic mode.

.. admonition:: Note to maintainers
   :class: note

   All methods that evaluate spectral quantities, conventionally named ``eval_*``,
   should accept a ``si: SpectralIndex`` parameter.

   If a new spectral mode is added, you need to:

   * add a new :class:`.SpectralIndex` subclass
   * add it to the :data:`.SPECTRAL_MODE_DISPATCH` dictionary, which maps
     spectral mode enum values to the corresponding spectral index type

   For experiments to support the newly added spectral mode, you will also need to
   update the ``eradiate/src/experiment`` package, in particular the
   ``_normalize_spectral()`` post-init method of the
   :class:`.EarthObservationExperiment`.
"""

from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod

import attrs
import pint
import pinttr

from .. import validators
from .._mode import ModeFlag, SubtypeDispatcher
from ..attrs import define, documented
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg


@attrs.define(eq=False, frozen=True, slots=True)
class SpectralIndex(ABC):
    """
    Abstract spectral index base class.

    See Also
    --------
    :class:`.MonoSpectralIndex`, :class:`.CKDSpectralIndex`
    """

    subtypes = SubtypeDispatcher("SpectralIndex")

    @property
    @abstractmethod
    def formatted_repr(self) -> str:
        """
        Formatted representation of the spectral index.
        """
        pass

    @property
    @abstractmethod
    def as_hashable(self) -> t.Hashable:
        """
        Hashable representation of the spectral index.

        Notes
        -----
        The hashable value is used by :class:`eradiate.experiments.Experiment`
        to identify the simulation results corresponding to a given spectral
        index.
        """
        pass

    @staticmethod
    def new(mode: ModeFlag | str | None = None, **kwargs) -> SpectralIndex:
        """
        Spectral index factory method.

        Parameters
        ----------
        mode : `.ModeFlag` or str, optional
            Spectral mode identifier. If ``None``, the current mode is used.

        Returns
        -------
        SpectralIndex
            Spectral index instance.

        Notes
        -----
        Create a new instance of the spectral index class corresponding to
        the specified mode.

        See Also
        --------
        :class:`.MonoSpectralIndex`, :class:`.CKDSpectralIndex`
        """
        if isinstance(mode, str):
            mode = mode.upper()
            if not mode.startswith("SPECTRAL_MODE_"):
                mode = f"SPECTRAL_MODE_{mode}"
            mode = ModeFlag[mode]
        si_cls = SpectralIndex.subtypes.resolve(mode)
        return si_cls(**kwargs)

    @staticmethod
    def from_dict(d: dict[str, t.Any]) -> SpectralIndex:
        d_copy = pinttr.interpret_units(d, ureg=ureg)
        return SpectralIndex.new(**d_copy)

    @staticmethod
    def convert(value: t.Any) -> SpectralIndex:
        if isinstance(value, SpectralIndex):
            return value
        elif isinstance(value, dict):
            return SpectralIndex.from_dict(value)
        else:
            raise ValueError(f"Cannot convert {value} to a spectral index.")


@SpectralIndex.subtypes.register(ModeFlag.SPECTRAL_MODE_MONO)
@define(eq=False, frozen=True, slots=True)
class MonoSpectralIndex(SpectralIndex):
    """
    Monochromatic spectral index.

    See Also
    --------
    :class:`CKDSpectralIndex`
    """

    w: pint.Quantity = documented(
        pinttr.field(
            default=550.0 * ureg.nm,
            units=ucc.deferred("wavelength"),
            on_setattr=None,
        ),
        doc="Wavelength.\n\nUnit-enabled field (default: ucc[wavelength]).",
        type="quantity",
        init_type="quantity or float",
        default="550.0 nm",
    )

    @w.validator
    def _w_validator(self, attribute, value):
        # wavelength must be a scalar quantity
        validators.on_quantity(validators.is_scalar)(self, attribute, value)

        # wavelength msut be positive
        validators.is_positive(self, attribute, value)

    @property
    def formatted_repr(self) -> str:
        return f"{self.w:g~P}"

    @property
    def as_hashable(self) -> float:
        return float(self.w.m_as(ureg.nm))


@SpectralIndex.subtypes.register(ModeFlag.SPECTRAL_MODE_CKD)
@define(eq=False, frozen=True, slots=True)
class CKDSpectralIndex(SpectralIndex):
    """
    CKD spectral index.

    See Also
    --------
    :class:`MonoSpectralIndex`
    """

    w: pint.Quantity = documented(
        pinttr.field(
            default=550.0 * ureg.nm,
            units=ucc.deferred("wavelength"),
            on_setattr=None,
        ),
        doc="Bin center wavelength.\n\nUnit-enabled field (default: ucc[wavelength]).",
        type="quantity",
        init_type="quantity or float",
        default="550.0 nm",
    )

    @w.validator
    def _w_validator(self, attribute, value):
        # wavelength must be a scalar quantity
        validators.on_quantity(validators.is_scalar)(self, attribute, value)

        # wavelength magnitude must be positive
        validators.is_positive(self, attribute, value.magnitude)

    g: float = documented(
        attrs.field(
            default=0.0,
            validator=attrs.validators.instance_of(float),
            converter=float,
        ),
        doc="g value.",
        type="float",
        init_type="float",
        default="0.0",
    )

    @g.validator
    def _g_validator(self, attribute, value):
        # g value must be between 0 and 1
        if not 0.0 <= value <= 1.0:
            raise ValueError(f"{attribute} must be between 0 and 1, got {value}")

    @property
    def formatted_repr(self) -> str:
        return f"{self.w:g~P}:{self.g:.3f}"

    @property
    def as_hashable(self) -> t.Tuple[float, float]:
        return (float(self.w.m_as(ureg.nm)), self.g)



from __future__ import annotations

import warnings

import attrs
import mitsuba as mi

from ._core import Surface
from ..bsdfs import BSDF, LambertianBSDF, bsdf_factory
from ..core import Ref, SceneTraversal, traverse
from ..shapes import RectangleShape, SphereShape, shape_factory
from ...attrs import define, documented
from ...exceptions import OverriddenValueWarning, TraversalError
from ...kernel import SceneParameter, SearchSceneParameter


@define(eq=False, slots=False)
class BasicSurface(Surface):
    """
    Basic surface [``basic``].

    A basic surface description consisting of a single shape and BSDF.
    """

    shape: None | RectangleShape | SphereShape = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(shape_factory.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of((RectangleShape, SphereShape))
            ),
        ),
        doc="Shape describing the surface. This parameter may be left unset "
        "for situations in which the task of setting its value is delegated to "
        "another component (*e.g.* an :class:`.Experiment` instance owning the "
        "surface object); however, if it is still unset upon kernel dictionary "
        "generation, the call to :meth:`.traverse` will raise a "
        ":class:`.TraversalError`.",
        type=".RectangleShape or .SphereShape or None",
        init_type=".RectangleShape or .SphereShape or dict, optional",
        default=":class:`.RectangleShape <RectangleShape()>`",
    )

    @shape.validator
    def _shape_validator(self, attribute, value):
        if value is not None and value.bsdf is not None:
            warnings.warn(
                f"while validating '{attribute.name}': "
                f"'{attribute.name}.bsdf' should be set to None; it will "
                "be overridden during kernel dictionary generation",
                OverriddenValueWarning,
            )

    bsdf: BSDF = documented(
        attrs.field(
            factory=LambertianBSDF,
            converter=bsdf_factory.convert,
            validator=attrs.validators.instance_of(BSDF),
        ),
        doc="The reflection model attached to the surface.",
        type=".BSDF",
        init_type=".BSDF or dict, optional",
        default=":class:`LambertianBSDF() <.LambertianBSDF>`",
    )

    def update(self) -> None:
        # Inherit docstring

        # Fix BSDF ID
        self.bsdf.id = self._bsdf_id

        # Force BSDF referencing if the shape is defined
        if self.shape is not None:
            if isinstance(self.shape.bsdf, BSDF):
                warnings.warn("Set BSDF will be overridden by surface BSDF settings.")
            self.shape.bsdf = Ref(id=self._bsdf_id)

    @property
    def _shape_id(self):
        """
        Mitsuba shape object identifier.
        """
        return f"{self.id}_shape"

    @property
    def _bsdf_id(self):
        """
        Mitsuba BSDF object identifier.
        """
        return f"{self.id}_bsdf"

    @property
    def _template_bsdfs(self) -> dict:
        kdict_template = traverse(self.bsdf)[0].data

        result = {}

        for key, param in kdict_template.items():
            result[f"{self._bsdf_id}.{key}"] = param

        return result

    @property
    def _template_shapes(self) -> dict:
        kdict_template = traverse(self.shape)[0].data

        result = {}

        for key, param in kdict_template.items():
            result[f"{self._shape_id}.{key}"] = param

        return result

    @property
    def _params_bsdfs(self) -> dict:
        umap_template = traverse(self.bsdf)[1].data

        result = {}

        for key, param in umap_template.items():
            # If no lookup strategy is set, we must add one
            if isinstance(param, SceneParameter) and param.search is None:
                param = attrs.evolve(
                    param,
                    search=SearchSceneParameter(
                        mi.BSDF,
                        self._bsdf_id,
                        parameter_relpath=key,
                    ),
                )

            result[f"{self._bsdf_id}.{key}"] = param

        return result

    @property
    def _params_shapes(self) -> dict:
        return {}

    def traverse(self, callback: SceneTraversal) -> None:
        # Inherit docstring
        if self.shape is None:
            raise TraversalError(
                "A 'BasicSurface' cannot be traversed if its 'shape' field is unset."
            )
        super().traverse(callback)



import subprocess


def get_command_output(cmd):
    """
    Run a command in the shell and return the output.
    Will raise and display any errors too.

    Parameters
    ----------
    cmd : str
        Command to run in shell, formatted as used in shell directly.

    Returns
    -------
    str
        Stdout of the process run in shell.
    """
    result = subprocess.run(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    stdout = result.stdout.decode().strip()
    if result.returncode == 0:
        return stdout

    stderr = result.stderr.decode().strip()
    raise RuntimeError(
        "Problem running '%s' (STDOUT: '%s' STDERR: '%s')" % (cmd, stdout, stderr)
    )


def get_repo_info():
    """
    Returns
    -------
    str
        The remote repo url.
    str
        The branch currently checked out.
    """
    out = get_command_output("git remote -v")
    repo = out.split("\n")[-1].split()[1]
    branch = get_command_output("git rev-parse --abbrev-ref HEAD")
    return (repo, branch)


def get_commit_info():
    """
    Get currently checked out git commit hash and time of commit.

    Returns
    -------
    str
        Commit hash currently checked out.
    str
        Time of commit.
    """
    commitHash = get_command_output("git rev-parse HEAD")
    commitTime = get_command_output("git log -n1 --pretty=%%ct %s" % commitHash)
    return (commitHash, str(int(commitTime) * 1000))



from . import atmosphere as atmosphere
from . import biosphere as biosphere
from . import bsdfs as bsdfs
from . import core as core
from . import geometry as geometry
from . import illumination as illumination
from . import integrators as integrators
from . import measure as measure
from . import phase as phase
from . import shapes as shapes
from . import spectra as spectra
from . import surface as surface



from ._core import Integrator as Integrator
from ._core import integrator_factory as integrator_factory
from ._path_tracers import PathIntegrator as PathIntegrator
from ._path_tracers import PiecewiseVolPathIntegrator as PiecewiseVolPathIntegrator
from ._path_tracers import VolPathIntegrator as VolPathIntegrator
from ._path_tracers import VolPathMISIntegrator as VolPathMISIntegrator



from __future__ import annotations

__all__ = [
    "auto_or",
    "convert_thermoprops",
    "on_quantity",
    "passthrough",
    "passthrough_type",
    "resolve_keyword",
    "resolve_path",
    "to_mi_scalar_transform",
]

import os
from pathlib import Path
from typing import Any, Callable

import attrs
import mitsuba as mi
import numpy as np
import pint
import xarray as xr

from .attrs import AUTO
from .data import fresolver
from .exceptions import DataError
from .typing import PathLike


def on_quantity(
    wrapped_converter: Callable[[Any], Any],
) -> Callable[[Any], Any]:
    """
    Apply a converter to the magnitude of a :class:`pint.Quantity`.

    Parameters
    ----------
    wrapped_converter : callable
        The converter which will be applied to the magnitude of a
        :class:`pint.Quantity`.

    Returns
    -------
    callable
    """

    def f(value: Any) -> Any:
        if isinstance(value, pint.Quantity):
            return wrapped_converter(value.magnitude) * value.units
        else:
            return wrapped_converter(value)

    return f


def auto_or(
    wrapped_converter: Callable[[Any], Any],
) -> Callable[[Any], Any]:
    """
    A converter that allows an attribute to be set to :data:`.AUTO`.

    Parameters
    ----------
    wrapped_converter : callable
        The converter that is used for non-:data:`.AUTO` values.

    Returns
    -------
    callable
    """

    def f(value):
        if value is AUTO:
            return value

        return wrapped_converter(value)

    return f


def passthrough(predicate: Callable[[Any], bool]) -> Callable[[Any], Any]:
    """
    Pass through values for which ``predicate`` returns ``True``; otherwise,
    apply wrapped converter.

    See Also
    --------
    :func:`.passthrough_type`
    """

    def wrapped_converter(converter: Callable | attrs.Converter) -> Any:
        if isinstance(
            converter, attrs.Converter
        ):  # See https://github.com/python-attrs/attrs/pull/1372

            def passthrough_converter(val, inst, field):
                return val if predicate(val) else converter(val, inst, field)

            return attrs.Converter(
                passthrough_converter, takes_self=True, takes_field=True
            )

        else:

            def passthrough_converter(val):
                return val if predicate(val) else converter(val)

            return passthrough_converter

    return wrapped_converter


def passthrough_type(types: type | tuple[type, ...]) -> Callable:
    """
    Pass through values of a specified type; otherwise, apply wrapped converter.

    See Also
    --------
    :func:`.passthrough`
    """
    return passthrough(lambda x: isinstance(x, types))


def resolve_keyword(path_forming_func: Callable[[Any], PathLike]) -> Callable:
    """
    Attempt resolving a keyword into a path constructed from a keyword by the
    ``path_forming_func`` parameter.

    If the generated path points to a file, the path is returned; otherwise,
    ``value`` is returned without modification.

    Parameters
    ----------
    path_forming_func : callable
        A callable with signature ``f(x: str) -> Path`` that constructs relative
        or absolute paths from keywords. Relative paths are then resolved by the
        file resolver.
    """

    def resolve_keyword_converter(value):
        path = fresolver.resolve(path_forming_func(value))
        if path.is_file():
            return path
        else:
            return value

    return resolve_keyword_converter


def resolve_path(value: PathLike) -> Path:
    """
    Resolve a file path with the file resolver. The current working directory is
    included in the path lookup.
    """
    return fresolver.resolve(value, cwd=True)


def load_dataset(value: PathLike) -> xr.Dataset:
    """
    Attempt loading a dataset given a path. If the path is relative, it is
    resolved by the file resolver first.

    Parameters
    ----------
    value
        Path to the targeted dataset.

    Raises
    ------
    DataError
        If the file could not be loaded.
    """
    path = resolve_path(value)
    try:
        return xr.load_dataset(path)
    except Exception as e:
        raise DataError(f"could not load dataset '{value}'") from e


def to_mi_scalar_transform(value: Any):
    """
    Convert an array-like value to a :class:`mitsuba.ScalarTransform4f`.
    If `value` is a Numpy array, it is used to initialize a
    :class:`mitsuba.ScalarTransform4f` without copy; if it is a list, a Numpy
    array is first created from it. Otherwise, `value` is forwarded without
    change.
    """
    if isinstance(value, np.ndarray):
        return mi.ScalarTransform4f(value)

    elif isinstance(value, list):
        return mi.ScalarTransform4f(np.array(value))

    else:
        return value


def convert_thermoprops(value: Any) -> xr.Dataset:
    """Converter for atmosphere thermophysical properties specifications."""
    import joseki

    # Dataset: do nothing
    if isinstance(value, xr.Dataset):
        return value

    # PathLike: try to load dataset
    if isinstance(value, (os.PathLike, str)):
        path = fresolver.resolve(value)
        if path.is_file():
            return joseki.load_dataset(path)
        else:
            raise ValueError(
                f"invalid path for 'thermoprops': {path} (expected a file)"
            )

    # Dictionary: forward to joseki.make()
    if isinstance(value, dict):
        return joseki.make(**value)

    # Anything else: raise error
    raise TypeError(
        f"invalid type for 'thermoprops': {type(value)} (expected Dataset or path-like)"
    )



import joseki
import numpy as np

from ... import fresolver
from ... import unit_registry as ureg
from ...experiments import AtmosphereExperiment

WAVELENGTH = [412, 443, 550, 670, 865, 1020, 1600, 2200]
ETA = [1.349303, 1.346833, 1.341266, 1.337636, 1.336949, 1.336949, 1.336949, 1.336949]
WB_COASTAL = [
    3.4678e-02,
    4.1939e-02,
    6.0228e-02,
    5.7141e-02,
    0.0000e00,
    0.0000e00,
    0.0000e00,
    0.0000e00,
]
WB_OPEN = [
    6.7215e-02,
    6.5480e-02,
    4.4756e-02,
    1.7900e-02,
    0.0000e00,
    0.0000e00,
    0.0000e00,
    0.0000e00,
]


def ocean_grasp_wavelength():
    return WAVELENGTH


def create_ocean_grasp(water_body_reflectance, wind_speed, has_atmoshphere=False):
    """
    Create the reference scene of the 3DREAMS project with ocean surface.

    Parameters
    ----------
    water_body_reflectance : list {WB_COASTAL, WB_OPEN}
        The water body reflectance spectrum. Should either be the WB_COASTAL or
        WB_OPEN initialized above.

    wind_speed: float
        Wind speed at mast height in m/s.

    has_atmosphere: bool
        Flags whether to include an atmosphere and particle layer as described
        by 3DREAMS scenario REF_OO_UB01_I_S20_PPL.

    Returns
    -------
    AtmosphereExperiment
        The atmosphere experiment corresponding to the GRASP scenario.
    """

    if has_atmoshphere:
        UB = fresolver.load_dataset(
            "tests/regression_test_specifications/ocean_grasp/REF_UB.nc"
        )

        spp = 10000
        atmosphere = {
            "type": "heterogeneous",
            "molecular_atmosphere": {
                "type": "molecular",
                "has_absorption": False,
                "thermoprops": joseki.make(
                    identifier="afgl_1986-us_standard",
                    z=np.arange(0, 40 + 0.1, 0.1) * ureg.km,
                ).joseki.rescale_to({"CO2": 360 * ureg.ppm}),
                "rayleigh_depolarization": 0.0,
            },
            "particle_layers": [
                {
                    "type": "particle_layer",
                    "bottom": 0 * ureg.km,
                    "top": 40 * ureg.km,
                    "distribution": {
                        "type": "exponential",
                        "rate": 40,
                    },
                    "tau_ref": 0.1,
                    "w_ref": 550 * ureg.nm,
                    "has_absorption": True,
                    "dataset": UB,
                }
            ],
        }
    else:
        spp = 1
        atmosphere = None

    config = {
        "surface": {
            "type": "ocean_grasp",
            "wind_speed": wind_speed,
            "eta": {
                "type": "interpolated",
                "wavelengths": WAVELENGTH,
                "values": ETA,
            },
            "water_body_reflectance": {
                "type": "interpolated",
                "wavelengths": WAVELENGTH,
                "values": water_body_reflectance,
            },
        },
        "illumination": {
            "zenith": 20.0,
            "zenith_units": "degree",
            "azimuth": 0.0,
            "azimuth_units": "degree",
            "type": "directional",
        },
        "measures": [
            {
                "type": "mdistant",
                "construct": "hplane",
                "zeniths": np.arange(-60, 61, 5),
                "zeniths_units": "degree",
                "azimuth": 0.0,
                "azimuth_units": "degree",
                "srf": {
                    "type": "multi_delta",
                    "wavelengths": WAVELENGTH,
                },
                "spp": spp,
            }
        ],
        "geometry": {
            "type": "plane_parallel",
            "zgrid": np.arange(0, 40 + 0.1, 0.1) * ureg.km,
            "toa_altitude": 40 * ureg.km,
        },
        "atmosphere": atmosphere,
        "integrator": {"type": "piecewise_volpath", "moment": True}
        if has_atmoshphere
        else {"type": "volpath", "moment": True},
    }

    return AtmosphereExperiment(**config)


def create_ocean_grasp_coastal_no_atm():
    """
    Ocean GRASP Coastal with no atmosphere
    ======================================

    This test case includes a coastal oceanic surface with no atmosphere.

    Parameters
    ----------

    * Surface: Square surface with Ocean GRASP BSDF with wind speed of 2 m/s and
    an coastal ocean water body reflectance spectrum.
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Multi Distant measure over the principal plane from -60 to 60 degrees
    with 5 degree steps.
    """
    return create_ocean_grasp(WB_COASTAL, 2.0, False)


def create_ocean_grasp_open_no_atm():
    """
    Ocean GRASP Open with no atmosphere
    ===================================

    This test case includes an open oceanic surface with no atmosphere.

    Parameters
    ----------

    * Surface: Square surface with Ocean GRASP BSDF with wind speed of 10 m/s and
    an open ocean water body reflectance spectrum.
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Multi Distant measure over the principal plane from -60 to 60 degrees
    with 5 degree steps.
    """
    return create_ocean_grasp(WB_OPEN, 10.0, False)


def create_ocean_grasp_open_atm():
    """
    Ocean GRASP Open with atmosphere
    ===================================

    This test case includes an open oceanic surface with atmosphere.

    Parameters
    ----------

    * Surface: Square surface with Ocean GRASP BSDF with wind speed of 10 m/s and
    an open ocean water body reflectance spectrum.
    * Atmosphere (if `has_atmosphere=True`): afgl US standard atmosphere with
    rescaled CO2 concentration to 360 ppm and depolarization set to zero.
    * Particle Layer (if `has_atmosphere=True): Exponentially distributed layer
    with absorption and optical depth of 0.1, distribution
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Multi Distant measure over the principal plane from -60 to 60 degrees
    with 5 degree steps.
    """
    return create_ocean_grasp(WB_OPEN, 10.0, True)



from __future__ import annotations

import attrs
import pint
import pinttrs

from ._core import BSDF
from ...attrs import define, documented
from ...kernel import DictParameter, SceneParameter
from ...units import unit_context_config as ucc
from ...units import unit_registry as ureg
from ...validators import is_positive


@define(eq=False, slots=False)
class OceanLegacyBSDF(BSDF):
    """
    Ocean Legacy BSDF [``ocean_legacy``].

    This BSDF implements the 6SV ocean surface model as described in
    :cite:t:`Kotchenova2006The6SVRTM`. This model treats the ocean as an opaque
    surface, and models the sunglint, whitecap and underlight components
    of the ocean reflectance. It depends on wind properties and
    pigmentation and chlorinity which makes it suitable to represent
    case I waters as defined by :cite:t:`Morel1988ModelingUpperOcean`.

    See Also
    --------
    :ref:`plugin-bsdf-ocean_legacy`

    Notes
    -----
     The ``wind_direction`` parameter indicates the azimuth angle of the wind and
    is interpreted using the :ref:`North left convention <sec-user_guide-conventions-azimuth>`.
    """

    wind_speed: pint.Quantity = documented(
        pinttrs.field(
            units=ureg("m/s").units,
            factory=lambda: 0.01 * ureg("m/s"),
            validator=[is_positive, pinttrs.validators.has_compatible_units],
        ),
        doc="Wind speed [m/s] at 10 meters above the surface.",
        type="quantity",
        init_type="quantity or float",
        default="0.01 m/s",
    )

    wind_direction: pint.Quantity = documented(
        pinttrs.field(
            factory=lambda: 0.0 * ureg.deg,
            validator=[is_positive, pinttrs.validators.has_compatible_units],
            units=ucc.deferred("angle"),
        ),
        doc="Wind azimuthal angle in *North Left-Hand convention*.\n\nUnit-enabled field (default units: ucc['angle']).",
        type="quantity",
        init_type="quantity or float",
        default="0.0 deg",
    )

    chlorinity: pint.Quantity = documented(
        pinttrs.field(
            units=ureg.Unit("g/kg"),
            factory=lambda: 19.0 * ureg("g/kg"),
            validator=[is_positive, pinttrs.validators.has_compatible_units],
        ),
        doc="Chlorinity of water.",
        type="quantity",
        init_type="quantity or float",
        default="19.0 g/kg",
    )

    pigmentation: pint.Quantity = documented(
        pinttrs.field(
            units=ureg.Unit("mg/m^3"),
            factory=lambda: 0.3 * ureg("mg/m^3"),
            validator=[is_positive, pinttrs.validators.has_compatible_units],
        ),
        doc="Pigmentation of water.",
        type="quantity",
        init_type="quantity or float",
        default="0.3 mg/m^3",
    )

    shadowing: bool = documented(
        attrs.field(
            converter=bool,
            validator=attrs.validators.instance_of(bool),
            default=True,
        ),
        doc="Indicates whether evaluation of BRDF computes shadowing and masking.",
        type="bool",
        default="True",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = {
            "type": "ocean_legacy",
            "wavelength": DictParameter(lambda ctx: ctx.si.w.m_as("nm")),
            "wind_speed": self.wind_speed.m_as("m/s"),
            "wind_direction": self.wind_direction.m_as("deg"),
            "chlorinity": self.chlorinity.m_as("g/kg"),
            "pigmentation": self.pigmentation.m_as("mg/m^3"),
            "shadowing": self.shadowing,
        }

        if self.id is not None:
            result["id"] = self.id

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        result = {"wavelength": SceneParameter(lambda ctx: ctx.si.w.m_as("nm"))}

        return result



"""Spectral response function (SRF) filtering algorithms."""

from __future__ import annotations

import datetime
import warnings
from pathlib import Path
from typing import Literal

import attrs
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pint
import rich
import scipy.integrate as spint
import xarray as xr
from rich.prompt import Confirm
from rich.table import Table

from . import converters
from ._version import _version as __version__
from .typing import PathLike
from .units import to_quantity
from .units import unit_registry as ureg
from .util.deprecation import deprecated

_trapezoid = np.trapezoid if int(np.__version__.split(".")[0]) >= 2 else np.trapz

convert_no_id = converters.passthrough_type(xr.Dataset)(converters.resolve_path)
convert = converters.passthrough_type(xr.Dataset)(
    attrs.converters.pipe(
        converters.resolve_keyword(lambda x: f"srf/{x}.nc"),
        converters.resolve_path,
    )
)


def update_attrs(srf: xr.Dataset, filter_name: str, filter_attr: str) -> None:
    """
    Update data set attributes to indicate that a filtering operation occurred.

    Parameters
    ----------
    srf : Dataset
        Data set whose attributes to update.

    filter_name : str
        Filter name.

    filter_attr : str
        Content of the 'filter' attribute.

    Notes
    -----
    * The update is done inplace.
    * If the dataset has no ``filter`` attribute, it is created. Otherwise, a
      new line is appended to it.
    """
    # filter attribute
    _value = srf.attrs.get("filter")
    previous_filter = _value + "\n" if _value is not None else ""

    # history attribute
    previous_history = srf.attrs["history"] + "\n"
    utcnow = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    author = f"eradiate {__version__}"
    history_attr = f"{utcnow} - data set filtering ({filter_name}) - {author}"

    # update attributes
    srf.attrs.update(
        {
            "history": f"{previous_history}{history_attr}",
            "filter": f"{previous_filter}{filter_attr}",
        }
    )


def wavelength_range_width(srf: PathLike | xr.Dataset) -> pint.Quantity:
    """
    Compute the wavelength range width of a spectral response function.

    Parameters
    ----------
    srf : path-like, Dataset
        Spectral response function data set.

    Notes
    -----
    The wavelength range width is defined as the difference between the upper
    and lower wavelength bounds.
    """
    srf = convert_no_id(srf)
    return to_quantity(srf.w).max() - to_quantity(srf.w).min()


def wavelength_bandwidth(srf: PathLike | xr.Dataset) -> pint.Quantity:
    r"""
    Compute the wavelength bandwidth of a spectral response function.

    Parameters
    ----------
    srf : path-like, Dataset
        Spectral response function data set.

    Notes
    -----
    The wavelength bandwidth is defined according to the following formula:

    .. math::

        \Delta \lambda = \int \lambda \, \phi(\lambda) \mathrm{d} \lambda

    where
        * :math:`\phi(\lambda)` is the spectral response function.
        * :math:`\lambda` is the wavelength.
        * :math:`\Delta \lambda` is the wavelength bandwidth.

    and the integral is performed over the wavelength range of the spectral
    response function.
    """
    srf = convert_no_id(srf)
    return srf.srf.integrate(coord="w").values * ureg(srf.w.attrs["units"])


def mean_wavelength(srf: PathLike | xr.Dataset) -> pint.Quantity:
    r"""
    Compute the mean wavelength of a spectral response function.

    Parameters
    ----------
    srf : path-like, Dataset
        Spectral response function data set.

    Notes
    -----
    The mean wavelength is defined according to the following formula:

    .. math::

        \overline{\lambda} = \frac{
            \int \lambda \, \phi(\lambda) \, \mathrm{d} \lambda
        }{
            \int \phi(\lambda) \, \mathrm{d} \lambda
        }

    where
        * :math:`\phi(\lambda)` is the spectral response function.
        * :math:`\lambda` is the wavelength.
        * :math:`\overline{\lambda}` is the mean wavelength.

    and the integrals are performed over the wavelength range of the spectral
    response function.
    """
    srf = convert_no_id(srf)
    srf_integrated = srf.srf.integrate(coord="w")
    mean_value = (srf.w * srf.srf).integrate(coord="w").values / srf_integrated.values
    return mean_value * ureg(srf.w.attrs["units"])


def filtering_summary(
    srf: PathLike | xr.Dataset, filtered: xr.Dataset
) -> dict[str, dict[str, int | pint.Quantity]]:
    srf = convert_no_id(srf)
    ni = srf.w.size
    nf = filtered.w.size
    w_min_i = to_quantity(srf.w).min()
    w_max_i = to_quantity(srf.w).max()
    w_min_f = to_quantity(filtered.w).min()
    w_max_f = to_quantity(filtered.w).max()
    range_width_i = wavelength_range_width(srf)
    range_width_f = wavelength_range_width(filtered)
    bandwidth_i = wavelength_bandwidth(srf)
    bandwidth_f = wavelength_bandwidth(filtered)
    mean_wavelength_i = mean_wavelength(srf)
    mean_wavelength_f = mean_wavelength(filtered)

    return {
        "Lower wavelength": {
            "Initial": f"{w_min_i:.1f~}",
            "Final": f"{w_min_f:.1f~}",
            "Difference": f"{(w_min_f - w_min_i):.1f~}",
        },
        "Upper wavelength": {
            "Initial": f"{w_max_i:.1f~}",
            "Final": f"{w_max_f:.1f~}",
            "Difference": f"{(w_max_f - w_max_i):.1f~}",
        },
        "# wavelength": {
            "Initial": f"{ni}",
            "Final": f"{nf}",
            "Difference": f"{(nf - ni)}",
        },
        "Wavelength range width": {
            "Initial": f"{range_width_i:.1f~}",
            "Final": f"{range_width_f:.1f~}",
            "Difference": f"{(range_width_f - range_width_i):.1f~}",
        },
        "Wavelength bandwidth": {
            "Initial": f"{bandwidth_i:.1f~}",
            "Final": f"{bandwidth_f:.1f~}",
            "Difference": f"{(bandwidth_f - bandwidth_i):.1f~}",
        },
        "Mean wavelength": {
            "Initial": f"{mean_wavelength_i:.1f~}",
            "Final": f"{mean_wavelength_f:.1f~}",
            "Difference": f"{(mean_wavelength_f - mean_wavelength_i):.1f~}",
        },
    }


def summarize(
    srf: PathLike | xr.Dataset,
    filtered: PathLike | xr.Dataset,
) -> Table:
    """
    Produce a summary table of the spectral response function filtering operation.

    Parameters
    ----------
    srf : Dataset
        Initial spectral response function data set.

    filtered : Dataset
        Filtered spectral response function data set.

    Returns
    -------
    :class:`~rich.table.Table`
        Summary table.
    """
    # convert inputs
    srf = convert_no_id(srf)
    filtered = convert_no_id(filtered)

    summary = filtering_summary(srf=srf, filtered=filtered)
    table = Table(
        title="Filtering summary",
        box=rich.box.HEAVY_HEAD,
    )

    columns = {
        "Characteristic": "left",
        "Initial": "center",
        "Final": "center",
        "Difference": "center",
    }

    for name, justification in columns.items():
        table.add_column(name, justify=justification)

    for key in summary:
        row_items = [key] + list(summary[key].values())

        if key == "Difference":
            pass
        else:
            table.add_row(*row_items)

    return table


def trim(srf: PathLike | xr.Dataset) -> xr.Dataset:
    """
    Trim all leading zeros except last and all trailing zeros except first.

    Parameters
    ----------
    srf:  path-like, Dataset
        Data set to trim.

    Returns
    -------
    Dataset
        Trimmed data set.
    """
    ds = convert_no_id(srf)

    # trim
    wsize = ds.srf.values.size
    fsize = np.trim_zeros(ds.srf.values, trim="f").size
    bsize = np.trim_zeros(ds.srf.values, trim="b").size
    istart = wsize - fsize - 1 if wsize > fsize else 0
    istop = bsize if bsize < wsize else wsize - 1
    trimmed = ds.isel(w=range(istart, istop + 1))

    # update history attribute
    previous_history = ds.attrs["history"] + "\n"
    utcnow = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    author = f"eradiate, version {__version__}"
    history_attr = f"{utcnow} - trimmed data set - {author}"
    trimmed.attrs.update({"history": f"{previous_history}{history_attr}"})

    return trimmed


def save(
    ds: xr.Dataset, path: PathLike, verbose: bool = False, dry_run: bool = False
) -> None:
    """
    Save dataset to disk.

    Parameters
    ----------
    ds : Dataset
        Dataset to save.

    path : path-like
        Path to which to save the dataset.

    verbose : bool
        If ``True``, display where the dataset is saved.

    dry_run : bool
        If ``True``, display where the dataset would be saved but does not save
        it.
    """
    output_path = Path(path).absolute()
    if dry_run:
        rich.print(f"Would write filtered data to {output_path}")
    else:
        if verbose:
            rich.print(f"Writing filtered data to {output_path}")
        ds.to_netcdf(path)


def trim_and_save(
    srf: PathLike | xr.Dataset,
    path: PathLike,
    verbose=False,
    show_plot: bool = False,
    dry_run: bool = False,
    interactive: bool = False,
) -> None:
    """
    Wraps around :meth:`trim` to save the filtered dataset.

    Parameters
    ----------
    srf: path-like, Dataset
        Data set to trim.

    path: path-like
        Path to which to save the filtered dataset.

    verbose : bool
        If ``True``, display a summary of the trimming operation.

    show_plot: bool
        If ``True``, display a figure illustrating the filtered region.

    dry_run: bool
        If ``True``, displays where the trimmed data set would be saved but
        does not save it.

    interactive: bool
        If ``True``, prompt the user to proceed to saving the filtered dataset.
        This is useful in combination with ``verbose=True`` and
        ``show_plot=True``.

    See Also
    --------
    :meth:`trim`
    """
    # convert inputs
    ds = convert_no_id(srf)
    output_path = Path(path).absolute()

    # trim
    trimmed = trim(srf=srf)

    # print trimming summary table
    if verbose:
        table = summarize(srf=ds, filtered=trimmed)
        rich.print(table)

    # save trimmed dataset
    if show_plot:
        show(
            ds=ds,
            title=" ".join(
                [srf.attrs["platform"], srf.attrs["instrument"], srf.attrs["band"]]
            ),
            trim_prior=False,
            threshold=None,
            wmin=to_quantity(trimmed.w).min(),
            wmax=to_quantity(trimmed.w).max(),
            percentage=None,
        )

    if interactive:
        if not Confirm.ask("Save filtered dataset?"):
            return

    save(
        ds=trimmed,
        path=output_path,
        verbose=verbose,
        dry_run=dry_run,
    )


def spectral_filter(
    srf: PathLike | xr.Dataset,
    wmin: pint.Quantity | None = None,
    wmax: pint.Quantity | None = None,
) -> xr.Dataset:
    """
    Drop points falling out of wavelength range specified by ``wmin`` and ``wmax``.

    Parameters
    ----------
    srf : path-like or Dataset
        Spectral response function data set to filter.

    wmin: quantity
        Minimum wavelength.

    wmax: quantity
        Maximum wavelength.

    Returns
    -------
    Dataset
        Filtered data set.
    """
    srf = convert_no_id(srf)

    # filter
    w_units = srf.w.attrs["units"]

    if wmin is not None:
        _wmin = wmin.m_as(w_units)
        filtered = srf.where(srf.w >= _wmin, drop=True)
        _wmin = f"{_wmin:.2f}"
    else:
        filtered = srf
        _wmin = 0.0

    if wmax is not None:
        _wmax = wmax.m_as(w_units)
        filtered = filtered.where(srf.w <= _wmax, drop=True)
        _wmax = f"{_wmax:.2f}"
    else:
        _wmax = "∞"

    # update filtered data set attributes
    update_attrs(
        srf=filtered,
        filter_name="spectral filter",
        filter_attr=(
            f"All points in the original data set that fell out of the "
            f"wavelength range [{_wmin}, {_wmax}] {w_units} were dropped."
        ),
    )

    # sanity check
    if filtered.w.size == 0:
        raise ValueError(
            f"Filtering this data set with wmin = {wmin} and wmax = {wmax} "
            f"would result in empty data set."
        )

    return filtered


def threshold_filter(srf: PathLike | xr.Dataset, value: float = 1e-3) -> xr.Dataset:
    """
    Drop data points where response is smaller or equal than a threshold value.

    Parameters
    ----------
    srf : path-like or Dataset
        Spectral response function data set to filter.

    value : float
        Spectral response threshold value.

    Raises
    ------
    ValueError: if the threshold value is not in [0, 1[.

    Returns
    -------
    Dataset
        Filtered data set.
    """
    # Convert input
    srf = convert_no_id(srf)

    # validate input
    if value < 0.0 or value >= 1.0:
        raise ValueError(f"threshold value should be in [0, 1[ (got {value}).")

    # check that filtering does not disconnect the wavelength space
    filter_indices = np.where(srf.srf.values > value)[0]
    consecutive = np.arange(filter_indices[0], filter_indices[0] + filter_indices.size)
    if not np.all(filter_indices == consecutive):
        warnings.warn(
            f"Filtering this data set with threshold value of {value} would "
            "disconnect the wavelength space. You probably do not want that."
        )

    # filter
    filtered = srf.where(srf.srf > value, drop=True)

    # update filtered data set attributes
    update_attrs(
        srf=filtered,
        filter_name="threshold filter",
        filter_attr=(
            f"All points in the original data set where the spectral response "
            f"function evaluated to {value} or smaller were dropped."
        ),
    )

    # sanity check
    if filtered.w.size == 0:
        raise ValueError(
            f"Filtering this data set with threshold value of {value} would "
            f"result in empty data set."
        )

    return filtered


def _integral_filter_bounds_walk(
    x: npt.ArrayLike, y: npt.ArrayLike, fraction: float
) -> tuple[tuple[int, int], float]:
    # Compute CDF
    cdf = np.concatenate(([0], spint.cumulative_trapezoid(y, x)))
    cdf /= cdf.max()

    # Compute indexes that realize the target fraction
    i_left = np.argwhere(cdf < 0.5 * fraction).max()
    i_right = np.argwhere(cdf > 1.0 - (0.5 * fraction)).min()
    cs = cdf[i_right] - cdf[i_left]

    return (i_left, i_right), cs


def _integral_filter_bounds_symmetry(
    x: npt.ArrayLike, y: npt.ArrayLike, fraction: float
) -> tuple[tuple[int, int], float]:
    # Insert mean in x array
    xmean = _trapezoid(y * x, x) / _trapezoid(y, x)
    i_xmean = np.argwhere(x < xmean).max() + 1
    xext = np.insert(x, i_xmean, xmean)
    yext = np.insert(y, i_xmean, np.interp(xmean, x, y))

    # Compute CDF
    cdf = np.concatenate(([0], spint.cumulative_trapezoid(yext, xext)))
    cdf /= cdf.max()

    # Compute symmetric indexes for which the enclosed CDF reaches the target fraction
    i_max = (len(xext) - 1) // 2
    for i in range(i_max):
        i_left = max([i_xmean - i, 0])
        i_right = min([i_xmean + i, len(xext) - 1])
        cs = cdf[i_right] - cdf[i_left]
        if cs >= 1.0 - fraction:
            break

    return (i_left, i_right - 1), cs


def integral_filter(
    srf: PathLike | xr.Dataset,
    percentage: float = 99.0,
    method: Literal["symmetry", "walk"] = "symmetry",
) -> xr.Dataset:
    """
    Keep only data that contribute to the integrated spectral response value
    to the amount of the specified percentage.

    Parameters
    ----------
    srf : path-like or Dataset
        Spectral response function data set to filter.

    percentage : float
        Keep data that contribute to this percentage of the integrated spectral
        response.

    method : "symmetry" or "walk"
        Trimming bound definition method.
        The ``"symmetry"`` method ensures that the bounds are positioned
        symmetrically with respect to the mean wavelength of the SRF.
        The ``"walk"`` method uses the cumulative integral of the SRF to
        position bounds eagerly while walking the spectral dimension.

    Raises
    ------
    ValueError: if the percentage value is not in ]0, 100].

    Returns
    -------
    Dataset
        Filtered data set.
    """
    # Convert input
    ds = convert_no_id(srf)

    # Validate percentage
    if not 0.0 < percentage <= 100.0:
        raise ValueError(f"value must be within ]0, 100.0] (got {percentage})")
    fraction = 1.0 - percentage / 100.0

    # Compute bounds and apply filtering method
    w = ds.w.values
    values = ds.srf.values

    if method == "symmetry":
        (i_left, i_right), cs = _integral_filter_bounds_symmetry(w, values, fraction)
    elif method == "walk":
        (i_left, i_right), cs = _integral_filter_bounds_walk(w, values, fraction)
    else:
        raise ValueError(f"Unknown method '{method}'")

    wmin, wmax = w[[i_left, i_right]]
    filtered = srf.where((srf.w >= wmin) & (srf.w <= wmax), drop=True)

    # Update filtered data set attributes
    update_attrs(
        srf=filtered,
        filter_name="integral filter",
        filter_attr=(
            f"Data points that did not contribute to {percentage} % of the "
            f"integrated spectral response were dropped."
        ),
    )

    # Sanity check
    if filtered.w.size == 0:
        raise ValueError(
            f"Filtering this data set with {percentage=} "
            f"would result in empty data set."
        )

    return filtered


def pad_zeros(ds: xr.Dataset) -> xr.Dataset:
    """
    Pad an SRF dataset with zero values to the left and right.

    Parameters
    ----------
    ds : Dataset
        SRF dataset


    Returns
    -------
    Dataset
    """

    # Extend spectral dimension
    w = ds.w.values
    dw = np.diff(ds.w)
    w = np.concatenate(([w[0] - dw[0]], w, [w[-1] + dw[-1]]))

    # Pad data variables
    data = {
        "srf": np.concatenate(([0.0], ds["srf"].values, [0.0])),
        "srf_u": np.concatenate(([np.nan], ds["srf_u"].values, [np.nan])),
    }

    # Create result variable
    result = xr.Dataset(
        {
            "srf": ("w", data["srf"], ds["srf"].attrs),
            "srf_u": ("w", data["srf_u"], ds["srf_u"].attrs),
        },
        coords={"w": ("w", w, ds["w"].attrs)},
        attrs=ds.attrs,
    )

    # Update dataset attributes
    update_attrs(
        srf=result,
        filter_name="pad_zeros",
        filter_attr="Added leading and trailing zeros.",
    )

    return result


def show(
    ds: PathLike | xr.Dataset,
    trim_prior: bool = True,
    title: str | None = None,
    threshold: float | None = None,
    wmin: pint.Quantity | None = None,
    wmax: pint.Quantity | None = None,
    percentage: float | None = None,
) -> None:
    """
    Show filtered region on spectral response function plot.

    Parameters
    ----------
    ds : path-like, Dataset
        Spectral response function to be filtered.

    trim_prior : bool
        If ``True``, trim spectral response function prior to filter.

    title : str, optional
        Figure title.

    threshold : float, optional
        Threshold value for the threshold filter.

    wmin : quantity, optional
        Lower wavelength value for the spectral filter.

    wmax : quantity, optional
        Upper wavelength value for the spectral filter.

    percentage : float, optional
        Percentage value for the integral filter (will use the 'symmetry' method).

    Raises
    ------
    ValueError:
        If the threshold value is not in [0, 1[.
    """
    # convert input
    ds = convert_no_id(ds)

    # setup figure
    fig, ax = plt.subplots(1, 1, figsize=(6, 4))

    plt_params = {"lw": 0.6, "marker": ".", "markersize": 2, "yscale": "log"}

    # optionally trim
    if trim_prior:
        trimmed = trim(srf=ds)
    else:
        trimmed = ds
        plt_params.update({"yscale": "linear"})

    # plot
    trimmed.srf.plot(**plt_params)

    if threshold is not None:
        if threshold < 0.0 or threshold >= 1.0:
            raise ValueError(f"threshold value should be in [0, 1[ (got {threshold}).")

        plt.axhline(y=threshold, color="red", lw=0.5)
        # keep region
        ax.fill_between(
            trimmed.w.values,
            0,
            trimmed.srf.values,
            where=trimmed.srf.values > threshold,
            facecolor="green",
            alpha=0.1,
        )
        plt.axhline(y=threshold, color="red", lw=0.5)
        # drop region
        ax.fill_between(
            trimmed.w.values,
            0,
            trimmed.srf.values,
            where=trimmed.srf.values <= threshold,
            facecolor="red",
            alpha=0.1,
        )

    if wmin is not None or wmax is not None:
        if wmin is not None:
            _wmin_value = wmin.m_as(ds.w.attrs["units"])
            plt.axvline(x=_wmin_value, color="red", lw=0.5)
            # drop region
            ax.fill_between(
                trimmed.w.values,
                0,
                trimmed.srf.values,
                where=trimmed.w.values < _wmin_value,
                facecolor="red",
                alpha=0.1,
            )
        else:
            _wmin_value = 0.0

        if wmax is not None:
            _wmax_value = wmax.m_as(ds.w.attrs["units"])
            plt.axvline(x=_wmax_value, color="red", lw=0.5)
            # drop region
            ax.fill_between(
                trimmed.w.values,
                0,
                trimmed.srf.values,
                where=trimmed.w.values > _wmax_value,
                facecolor="red",
                alpha=0.1,
            )
        else:
            _wmax_value = np.inf

        # keep region
        ax.fill_between(
            trimmed.w.values,
            0,
            trimmed.srf.values,
            where=(
                (trimmed.w.values >= _wmin_value) & (trimmed.w.values <= _wmax_value)
            ),
            facecolor="green",
            alpha=0.1,
        )

    if percentage is not None:
        x = ureg.convert(ds.w.values, ds.w.units, "nm")
        y = ds.srf.values
        fraction = 1.0 - percentage / 100.0
        wmin, wmax = x[list(_integral_filter_bounds_symmetry(x, y, fraction)[0])]
        # drop regions
        plt.axvline(x=wmin, color="red", lw=0.5)
        ax.fill_between(
            trimmed.w.values,
            0,
            trimmed.srf.values,
            where=trimmed.w.values < wmin,
            facecolor="red",
            alpha=0.1,
        )
        plt.axvline(x=wmax, color="red", lw=0.5)
        ax.fill_between(
            trimmed.w.values,
            0,
            trimmed.srf.values,
            where=trimmed.w.values > wmax,
            facecolor="red",
            alpha=0.1,
        )
        # keep region
        ax.fill_between(
            trimmed.w.values,
            0,
            trimmed.srf.values,
            where=((trimmed.w.values >= wmin) & (trimmed.w.values <= wmax)),
            facecolor="green",
            alpha=0.1,
        )

    if title is not None:
        plt.title(title)

    plt.tight_layout()
    plt.grid()
    plt.show()


def filter_srf(
    srf: PathLike | xr.Dataset,
    path: PathLike,
    verbose: bool = False,
    show_plot: bool = False,
    dry_run: bool = False,
    interactive: bool = False,
    trim_prior: bool = True,
    threshold: float | None = None,
    wmin: pint.Quantity | None = None,
    wmax: pint.Quantity | None = None,
    percentage: float | None = None,
    pad: bool = False,
) -> xr.Dataset:
    """
    Filter a spectral response function data set.

    Parameters
    ----------
    srf : path-like, Dataset
        Spectral response function data set to filter.

    path : path-like
        Path to which to save the filtered data set.

    verbose : bool
        If ``True``, display a filtering summary table and the path to which
        filtered data set is saved.

    show_plot : bool
        If ``True``, display a figure emphasizing the filtered region.

    dry_run : bool
        If ``True``, display the path to which the filtered data set would be
        saved, but does not write the data set to the disk.

    interactive : bool
        If ``True``, prompt the user to proceed to saving the filtered dataset.
        This is useful in combination with ``verbose=True`` and
        ``show_plot=True``.

    trim_prior : bool
        Trim the data set prior to filtering.

    threshold : float, optional
        Threshold value for the threshold filter.
        See :meth:`threshold_filter`.
        If ``None``, disable the threshold filter.

    wmin : quantity, optional
        Lower wavelength bound for the spectral filter.
        See :meth:`spectral_filter`.
        If both ``wmin`` and ``wmax`` are ``None``, disable the spectral filter.

    wmax : quantity, optional
        Upper wavelength bound for the spectral filter.
        See :meth:`spectral_filter`.
        If both ``wmin`` and ``wmax`` are ``None``, disable the spectral filter.

    percentage : float, optional
        Percentage value for the integral filter.
        See :meth:`integral_filter`.
        If ``None``, disable the integral filter.

    pad : bool, default: False
        If True, pad SRF data with leading and trailing zeros.

    Returns
    -------
    filtered : Dataset
        Filtered dataset.

    Notes
    -----
    Select the filtering algorithms corresponding to the specified parameters.

    If multiple filters are selected, they are applied in this order:

    * integral filter
    * spectral filter
    * threshold filter

    See Also
    --------
    :meth:`integral_filter`, :meth:`spectral_filter`,
    :meth:`threshold_filter`
    """
    # convert inputs
    srf = convert_no_id(srf)
    output_path = Path(path).absolute()

    # trimming
    trimmed = trim(srf=srf) if trim_prior else srf

    filtered = trimmed

    # integral filter
    if percentage is not None:
        filtered = integral_filter(srf=filtered, percentage=percentage)

    # spectral filter
    if wmin is not None or wmax is not None:
        filtered = spectral_filter(srf=filtered, wmin=wmin, wmax=wmax)

    # threshold filter
    if threshold is not None:
        filtered = threshold_filter(srf=filtered, value=threshold)

    # pad with zeros
    if pad:
        filtered = pad_zeros(filtered)

    # print filtering summary table
    if verbose:
        table = summarize(srf=srf, filtered=filtered)
        rich.print(table)

    # save filtered dataset
    if show_plot:
        show(
            ds=trimmed,
            title=" ".join(
                [srf.attrs["platform"], srf.attrs["instrument"], srf.attrs["band"]]
            ),
            trim_prior=trim_prior,
            threshold=threshold,
            wmin=wmin,
            wmax=wmax,
            percentage=percentage,
        )

    if interactive:
        if not Confirm.ask("Save filtered dataset?"):
            return

    save(ds=filtered, path=output_path, verbose=verbose, dry_run=dry_run)

    return filtered


@deprecated(
    deprecated_in="0.29.2",
    removed_in="0.31.0",
    details="eradiate.srf_tools.make_gaussian is moved to "
    "eradiate.spectral.response.make_gaussian",
)
def make_gaussian(*args, **kwargs) -> xr.Dataset:
    """
    Generate a Gaussian spectral response function dataset from central
    wavelength and full width at half maximum values.

    Parameters
    ----------
    wl_center : quantity or float
        Central wavelength of the Gaussian distribution.
        If passed as a float, the value is interpreted as being given in nm.

    fwhm : quantity or float
        Full width at half maximum of the Gaussian distribution.
        If passed as a float, the value is interpreted as being given in nm.

    cutoff : float, default: 3.0
        Cut-off, in multiples of the standard deviation σ.

    wl : quantity or array-like, optional
        Mesh used to evaluate the discretized distribution. If unset, a regular
        mesh with spacing given by ``wl_res`` is used.
        If passed as an array, the value is interpreted as being given in nm.

    wl_res : quantity or float, optional
        Resolution of the automatic spectral mesh if relevant.
        If passed as a float, the value is interpreted as being given in nm.

    pad : bool, default: False
        If True, pad SRF data with leading and trailing zeros.

    normalize : bool, default: True
        If ``True``, the generated SRF data is normalized to have a maximum
        equal to 1.

    Returns
    -------
    Dataset
        A dataset compliant with the Eradiate SRF format. The uncertainty
        variable is set to NaN.
    """
    from eradiate.spectral.response import make_gaussian as _make_gaussian

    return _make_gaussian(*args, **kwargs)



"""
Extensions and helpers for tutorials (legacy).
"""

import warnings

from eradiate.tutorials import *  # noqa: F403

warnings.warn(
    DeprecationWarning(
        "The 'eradiate.notebook.tutorials' module is deprecated: its features "
        "have been moved to 'eradiate.tutorials'."
    )
)



import os
import platform
import subprocess

import click
import json5

import eradiate

from . import utils
from .asvdb import ASVDb


@click.command
@click.option(
    "--set-commit-hash",
    default=None,
    help="Set the commit hash to use when recording benchmark results. "
    "By default, will use the currently checked out environment for dev environments "
    "and the eradiate version tag's commit for prod environment. Note also "
    "that if git-ref is specified, this will overwrite it when recording the results.",
)
@click.option(
    "--archive-dir",
    default=None,
    help="Specify the directory to another ASV database in which the result should be archived. "
    "The path should point to the root of the database (where the asv.conf.json lives).",
)
# NM - 25/09/24: Disable ASV git cloning features broken by the migration to pixi.
# @click.option(
#     "--git-ref",
#     default=None,
#     help="Specifies a range at which benchmarks should be run. Note that this will run using ASV's "
#     "contained environments and clone and build the project at each commit in the range.",
# )
@click.option("-v", "--verbose", is_flag=True, help="increase verbosity")
@click.option(
    "-e",
    "--show-stderr",
    is_flag=True,
    help="Display the stderr output from the benchmarks.",
)
@click.option(
    "-q",
    "--quick",
    is_flag=True,
    help="Do a 'quick' run, where each benchmark function is run only once. "
    "This is useful to find basic errors in the benchmark "
    "functions faster. The results are unlikely to be useful, "
    "and thus are not saved.",
)
@click.option(
    "-m",
    "--machine",
    default=None,
    help="Use the given name to retrieve machine information. If not provided, "
    "the hostname is used. If no entry with that name is found, and there "
    "is only one entry in ~/.asv-machine.json, that one entry will be used.",
)
def benchmark(
    set_commit_hash, archive_dir, git_ref, verbose, show_stderr, quick, machine
):
    """
    Benchmark test suite based on ASV. By default, runs the test suite on the
    current environment. Specify a git-ref range to run the benchmark on
    cloned projects (ASV's default).

    Warning! Production environment cannot be tested yet because no
    packaged versions contain those changes yet.
    """

    # get current benchmark directory
    eradiate_dir = os.environ.get("ERADIATE_SOURCE_DIR")
    benchmark_dir = os.path.join(eradiate_dir, "benchmarks")
    conf_path = os.path.join(benchmark_dir, "asv.conf.json")
    branches = []
    conf = None
    with open(conf_path) as conf_file:
        conf = json5.load(conf_file)
        branches = conf["branches"]

    cmd = ["asv", "run"]

    # NM - 25/09/24: Disable ASV git cloning features broken by the migration to pixi.
    # if git_ref:
    #     cmd.append(git_ref)
    # else:

    cmd.append("-E")
    cmd.append("existing:same")

    # Pass arguments to ASV
    if verbose:
        cmd.append("-v")

    if show_stderr:
        cmd.append("-e")

    if quick:
        cmd.append("-q")

    if machine:
        cmd.append(f"-m={machine}")

    # By default, dev environment will use the checked out commit hash
    # and prod the commit corresponding to their version tag.
    # if set_commit_hash is None and git_ref is None:
    if set_commit_hash is None:
        version = eradiate.__version__
        if "dev" in version:
            set_commit_hash, _ = utils.get_commit_info()
        else:
            cmd = f"git rev-list -n 1 v{version}"
            if verbose:
                click.echo(f"Running {cmd}")
            set_commit_hash = utils.get_command_output(cmd)

    if set_commit_hash:
        cmd.append(f"--set-commit-hash={set_commit_hash}")

        commit_branches_raw = utils.get_command_output(
            f"git branch --contains {set_commit_hash} "
        )

        commit_branches_raw = commit_branches_raw.replace("*", "")
        commit_branches_raw = commit_branches_raw.replace(" ", "")
        commit_branches = commit_branches_raw.split("\n")

        branch_in_conf = False
        for branch in branches:
            if branch in commit_branches:
                branch_in_conf = True

        if not branch_in_conf:
            if click.confirm(
                f"The branch '{commit_branches[0]}' you are about to benchmark isn't tracked by asv.conf.json. Are you sure you want to continue?",
                abort=True,
            ):
                pass

    # construct and run the full command
    cmd = " ".join(cmd)
    if verbose:
        click.echo(f"Running : {cmd}")
    subprocess.run(cmd, shell=True)

    if archive_dir and not quick:
        if verbose:
            click.echo("Archiving results:")

        if machine is None:
            # get default machine name if none were given
            uname = platform.uname()
            machine = uname.node

        if verbose:
            click.echo(f"\t Loading current database at {benchmark_dir}")

        # get all results in database
        current_db = ASVDb(benchmark_dir)
        results = current_db.getResults()

        if verbose:
            click.echo(f"\t Loading target database at {archive_dir}")

        (repo, branch) = utils.get_repo_info()
        target_db = ASVDb(archive_dir, repo=repo, branches=[branch])

        commit_list = []

        if set_commit_hash:
            commit_list.append(set_commit_hash[:8])

        # NM - 25/09/24: Disable ASV git cloning features broken by the migration to pixi.
        # elif git_ref:
        #     commit_list_raw = utils.get_command_output(f"git rev-list {git_ref}")
        #     commit_list_raw = commit_list_raw.split("\n")
        #     commit_list = [commit[:8] for commit in commit_list_raw]

        # filter by machine name and commit hash
        for result in results:
            info = result[0]
            if info.machineName == machine and info.commitHash[:8] in commit_list:
                if verbose:
                    click.echo(
                        f"\t Archiving {info.commitHash[:8]}-{info.envName}.json"
                    )
                target_db.addResults(info, result[1])


if __name__ == "__main__":
    benchmark()



from __future__ import annotations

import typing as t
from numbers import Number

import attrs
import numpy as np

from .attrs import AUTO
from .units import PhysicalQuantity
from .units import unit_registry as ureg


def is_scalar(_, attribute, value):
    """
    Validate iff value is of scalar type in the sense of Numpy.

    Raises
    ------
    TypeError
        If the value is not scalar.
    """
    if not np.isscalar(value):
        raise TypeError(f"'{attribute.name}' must be scalar, got {value}")


def is_number(_, attribute, value):
    """
    Validate iff value is of a numeric type.

    Raises
    ------
    TypeError
        If the value is not a :class:`Number`.
    """
    if not isinstance(value, Number):
        raise TypeError(
            f"{attribute.name} must be a real number, "
            f"got {value} which is a {value.__class__}"
        )


def is_vector3(instance, attribute, value):
    """
    Validate iff value can be converted to a (3,) Numpy array.

    Raises
    ------
    TypeError
        If value cannot be converted to a (3,) :class:`numpy.ndarray`.
    """
    return attrs.validators.deep_iterable(
        member_validator=is_number, iterable_validator=has_len(3)
    )(instance, attribute, value)


def is_positive(_, attribute, value):
    """
    Validate iff value is a positive number.

    Raises
    ------
    ValueError
        If the value is not positive or zero.
    """
    if value < 0.0:
        raise ValueError(f"{attribute} must be positive or zero, got {value}")


def all_positive(_, attribute, value):
    """
    Validate iff value is a vector with all its values positive.

    Raises
    ------
    ValueError
        If not all values are positive.
    """
    if isinstance(value, ureg.Quantity):
        value = value.magnitude
    if np.any(np.array(value) < 0):
        raise ValueError(f"{attribute} must be all positive or zero, got {value}")


def all_strictly_positive(_, attribute, value):
    """
    Validate iff value is a vector with all its values strictly positive.

    Raises
    ------
    ValueError
        If not all values are strictly positive.
    """
    if isinstance(value, ureg.Quantity):
        value = value.magnitude
    if np.any(np.array(value) <= 0):
        raise ValueError(f"{attribute} must be all strictly positive, got {value}")


def path_exists(_, attribute, value):
    """
    Validate iff initializer is called with a value defining a path to an
    existing location.

    Raises
    ------
    FileNotFoundError
        If the value is not a :class:`pathlib.Path` which points to an existing
        location.
    """
    if not value.exists():
        raise FileNotFoundError(
            f"{attribute} points to '{str(value)}' (path does not exist)"
        )


def is_file(_, attribute, value):
    """
    Validate iff initializer is called with a value defining a path to an
    existing file.

    Raises
    ------
    FileNotFoundError
        If the value is not a :class:`pathlib.Path` which points to an existing
        file.
    """
    if not value.is_file():
        raise FileNotFoundError(
            f"{attribute.name} points to '{str(value)}' (not a file)"
        )


def is_dir(_, attribute, value):
    """
    Validate iff initializer is called with a value defining a path to an
    existing directory.

    Raises
    ------
    FileNotFoundError
        If the value is not a :class:`pathlib.Path` which points to an existing
        directory.
    """
    if not value.is_dir():
        raise FileNotFoundError(
            f"{attribute.name} points to '{str(value)}' (not a directory)"
        )


def is_sorted(
    order: t.Literal["ascending", "descending"] = "ascending", strict: bool = False
):
    """
    Validate iff a 1D array is sorted.

    Parameters
    ----------
    order : {"ascending", "descending"}, default: "ascending"
        Order expected for the array.

    strict : bool, default: False
        If ``True``, check for strict inequality.
    """
    if order == "ascending":
        if strict:

            def cmp(a):
                return np.all(a[:-1] < a[1:])
        else:

            def cmp(a):
                return np.all(a[:-1] <= a[1:])
    else:
        if strict:

            def cmp(a):
                return np.all(a[:-1] > a[1:])
        else:

            def cmp(a):
                return np.all(a[:-1] >= a[1:])

    def f(_, attribute, value):
        if not cmp(value):
            raise ValueError(f"while validating {attribute}: array must be sorted")

    return f


def has_len(size: int):
    """
    Validate iff value is a vector with specified length.

    Parameters
    ----------
    size : int
        Expected size of the validated value.

    Raises
    ------
    ValueError
        If the value does not have the expected size.
    """

    def f(_, attribute, value):
        if len(value) != size:
            raise ValueError(
                f"{attribute} must be have length {size}, "
                f"got {value} of length {len(value)}"
            )

    return f


def has_quantity(quantity: PhysicalQuantity | str | None):
    """
    Validate iff initializer is called with a value  featuring a ``quantity``
    field set to an expected value.

    Parameters
    ----------
    quantity : :class:`.PhysicalQuantity` or str or None
        Expected quantity field.

    Raises
    ------
    ValueError
        If the value's ``quantity`` field does not match the expected value.
    """
    if quantity is not None:
        quantity = PhysicalQuantity(quantity)

    def f(_, attribute, value):
        if value.quantity != quantity:
            raise ValueError(
                f"incompatible quantity '{value.quantity}' "
                f"used to set field '{attribute.name}' "
                f"(allowed: '{quantity}')"
            )

    return f


def on_quantity(wrapped_validator: t.Callable):
    """
    Apply a validator to the magnitude of a quantity.

    Parameters
    ----------
    wrapped_validator : callable
        The validator applied to the value's magnitude.
    """

    def f(instance, attribute, value):
        if isinstance(value, ureg.Quantity):
            return wrapped_validator(instance, attribute, value.magnitude)
        else:
            return wrapped_validator(instance, attribute, value)

    return f


def auto_or(*wrapped_validators):
    """
    Allow for an attribute to be set to :class:`.AUTO`.

    Parameters
    ----------
    *wrapped_validators : callable
        Validators to be applied to values not equal to :class:`.AUTO`.
    """

    def f(instance, attribute, value):
        if value is AUTO:
            return

        for validator in wrapped_validators:
            validator(instance, attribute, value)

    return f



"""
Extensions and helpers for tutorials.
"""

from __future__ import annotations

import numpy as np
import xarray as xr
from matplotlib import pyplot as plt
from numpy.typing import ArrayLike

from .config import settings
from .frame import AzimuthConvention
from .scenes.atmosphere import AbstractHeterogeneousAtmosphere
from .spectral import SpectralIndex


def plot_polarfilm(
    da: xr.DataArray,
    levels: int | ArrayLike = 16,
    show_contour: bool = True,
    show_azimuth: bool = False,
    figsize: tuple = (3, 3),
    azimuth_convention: AzimuthConvention | str | None = None,
    vmin: float | None = None,
    vmax: float | None = None,
    theta_max: float = 90.0,
):
    """
    Plot film data generated by a hemispherical distant sensor as a polar
    contour plot.

    Parameters
    ----------
    da : xarray.DataArray
        Data array to be plotted.

    levels : int or array-like, default: 16
        Number of levels on the contour plot.

    show_contour : bool, default: True
        If ``True``, show contour outlines.

    show_azimuth : bool, default: False
        [Debug option] If ``True``, show azimuth values as points.

    figsize : tuple, default: (3, 3)
        Figure size.

    azimuth_convention : .AzimuthConvention or str, optional
        Azimuth convention of plotted data (will adjust the azimuth of the plot).

    vmin, vmax : float, optional
        Range of values to plot. If unset, automatically from the data.

    theta_max : float, default: 90.0
        Maximum value of the azimuth angle on plots.

    Returns
    -------
    matplotlib.Figure
    """

    import matplotlib.pyplot as plt
    import matplotlib.tri as tri

    # Filter grazing angles
    da = (
        da.where(da.vza <= theta_max)
        .dropna("x_index", how="all")
        .dropna("y_index", how="all")
    )
    theta_max = float(da.vza.max())
    rscale = 90.0 / theta_max

    # Prepare data for plotting
    azimuth_convention = (
        AzimuthConvention.convert(azimuth_convention)
        if azimuth_convention is not None
        else settings.AZIMUTH_CONVENTION
    )
    values = da.transpose("x_index", "y_index").values.ravel()
    zeniths = da["vza"].values.ravel()  # Degree
    azimuths = np.deg2rad(da["vaa"].values).ravel()  # Radian

    # Create triangulation
    x = zeniths * np.cos(azimuths) * rscale
    y = zeniths * np.sin(azimuths) * rscale
    triangles = tri.Triangulation(x, y)

    # Make plot
    fig = plt.figure(0, figsize=figsize)
    rect = [0, 0, 1, 1]

    # -- Main plot in Cartesian coordinates
    ax_cartesian = fig.add_axes(rect, aspect="equal")
    ax_cartesian.axis("off")  # Hide axis
    ctr = ax_cartesian.tricontourf(
        triangles,
        values,
        levels=levels,
        cmap="viridis",
        vmin=vmin,
        vmax=vmax,
    )

    if show_contour:
        ax_cartesian.tricontour(
            triangles,
            values,
            levels=levels,
            linewidths=0.5,
            colors="w",
            alpha=0.5,
            vmin=vmin,
            vmax=vmax,
        )

    if show_azimuth:
        ax_cartesian.scatter(x, y, c=azimuths, cmap="plasma", s=3)

    ax_cartesian.set_xlim([-90, 90])  ## Match limits with the full zenith range
    ax_cartesian.set_ylim([-90, 90])

    ## Polar axes
    ax_polar = fig.add_axes(rect, polar=True, facecolor="none")
    ax_polar.set_rlim([0, theta_max])  # Cover the full zenith value range
    ax_polar.grid(False, axis="x")
    ax_polar.grid(axis="y", alpha=0.25)
    ax_polar.tick_params(axis="y", colors="white")
    ax_polar.set_rlabel_position(120)

    ax_polar.set_theta_offset(azimuth_convention.value[0])
    ax_polar.set_theta_direction(azimuth_convention.value[1])

    # Add the color bar (important: both axes must be adjusted)
    fig.colorbar(ctr, ax=[ax_cartesian, ax_polar], shrink=0.75, pad=0.15)

    return fig


def plot_sigma_t(
    *atmospheres: AbstractHeterogeneousAtmosphere,
    labels: list[str] | None = None,
    altitude_extent: tuple[float, float] | None = None,
    si: SpectralIndex | None = None,
    show: bool = True,
) -> tuple[plt.Figure, plt.Axes]:
    """
    Display the extinction coefficient of one or several atmosphere objects for
    a single spectral context.

    Parameters
    ----------
    *atmospheres : .AbstractHeterogeneousAtmosphere
        One or several atmosphere objects for which to plot the extinction
        coefficient.

    labels : list of strings, optional
        Labels associated with the passed atmosphere objects. If unset, no
        legend will be added to the plot.

    altitude_extent : tuple of float
        A (min, max) altitude pair (in km) to which the plot is restricted.

    si : .SpectralIndex, optional
        The spectral index at which the extinction coefficient is evaluated.
        If unset, a default spectral index is created using
        :meth:`.SpectralIndex.new`.

    show : bool, optional
        If ``True``, return ``None`` and display the plot. Otherwise, return a
        (``Figure``, ``Axes``) pair.
    """

    from matplotlib.ticker import ScalarFormatter

    from eradiate.units import to_quantity

    if si is None:  # If none is found, fall back to the default (550 nm)
        si = SpectralIndex.new()
    else:  # Otherwise, apply conversion protocol
        si = SpectralIndex.convert(si)

    if labels is None:
        label_iter = iter([None for _ in atmospheres])
    else:
        label_iter = iter(labels)

    fig, ax = plt.subplots(1, 1)

    with plt.rc_context({"lines.linestyle": ":", "lines.marker": "."}):
        for atmosphere in atmospheres:
            altitude = to_quantity(atmosphere.eval_radprops(si=si).z_layer).m_as("km")
            sigma_t = to_quantity(atmosphere.eval_radprops(si=si).sigma_t).m_as("1/m")
            ax.plot(altitude, sigma_t, label=next(label_iter))

    formatter = ScalarFormatter(useMathText=True)
    formatter.set_powerlimits((-3, 2))
    ax.yaxis.set_major_formatter(formatter)
    ax.set_ylabel("Extinction coefficient [1/m]")

    if labels is not None:
        fig.legend(
            bbox_to_anchor=(1.0, 0.5),
            loc="center left",
            borderaxespad=0.0,
        )

    if altitude_extent is not None:
        ax.set_xlim(altitude_extent)
    ax.set_xlabel("Altitude [km]")

    plt.tight_layout()
    if show:
        plt.show()
    else:
        return fig, ax


def load_ipython_extension(ipython):
    """
    IPython extension for Eradiate tutorials. Should be loaded at the top of the
    tutorial notebook.

    It sets the Matplotlib style and prints the current date and Eradiate
    version as markdown.

    See Also
    --------
    :func:`eradiate.plot.set_style`

    Notes
    -----

    This extension should be loaded using the IPython magic:

    .. code::

       %load_ext eradiate.tutorials
    """

    import datetime

    from IPython.display import Markdown, display

    from . import __version__
    from .plot import set_style

    set_style()
    xr.set_options(display_expand_data=False)
    display(
        Markdown(
            f"*Last updated: {datetime.datetime.now():%Y-%m-%d %H:%M} "
            f"(eradiate v{__version__})*"
        )
    )



from __future__ import annotations

import mitsuba as mi
import numpy as np
import pint
import pinttr
from pinttr.util import ensure_units

from ._core import ShapeNode
from ..bsdfs import BSDF
from ..core import BoundingBox
from ...attrs import define, documented
from ...constants import EARTH_RADIUS
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


def _normalize(v: np.typing.ArrayLike) -> np.ndarray:
    return np.array(v) / np.linalg.norm(v)


@define(eq=False, slots=False)
class SphereShape(ShapeNode):
    """
    Sphere shape [``sphere``].

    This shape represents a sphere parametrized by its centre and radius.

    Notes
    -----
    * If the `to_world` parameter is set, it will be appended to the position
      and scaling defined by the `center` and `radius` parameters.
    """

    center: pint.Quantity = documented(
        pinttr.field(factory=lambda: [0, 0, 0], units=ucc.deferred("length")),
        doc="Location of the centre of the sphere. Unit-enabled field "
        "(default: ``ucc['length']``).",
        type="quantity",
        init_type="quantity or array-like, optional",
        default="[0, 0, 0]",
    )

    radius: pint.Quantity = documented(
        pinttr.field(
            factory=lambda: 1.0 * ucc.get("length"), units=ucc.deferred("length")
        ),
        doc="Sphere radius. Unit-enabled field (default: ``ucc['length']``).",
        type="quantity",
        init_type="quantity or float, optional",
        default="1.0",
    )

    @property
    def bbox(self) -> BoundingBox:
        length_units = ucc.get("length")
        if self.to_world is not None:
            trafo = (
                self.to_world
                @ mi.Transform4f.translate(self.center.m_as(length_units))
                @ mi.Transform4f.scale(self.radius.m_as(length_units))
            )
        else:
            trafo = mi.Transform4f.translate(
                self.center.m_as(length_units)
            ) @ mi.Transform4f.scale(self.radius.m_as(length_units))

        c = trafo @ (0, 0, 0)
        r = np.linalg.norm(trafo @ (1, 0, 0) - c)

        p1 = pint.Quantity(np.array(c + np.array((-1, -1, -1)) * r), length_units)
        p2 = pint.Quantity(np.array(c + np.array((1, 1, 1)) * r), length_units)

        return BoundingBox(p1, p2)

    @property
    def template(self) -> dict:
        result = {
            "type": "sphere",
            "center": self.center.m_as(uck.get("length")),
            "radius": self.radius.m_as(uck.get("length")),
        }
        if self.to_world is not None:
            result["to_world"] = self.to_world
        return result

    def contains(self, p: np.typing.ArrayLike, strict: bool = False) -> bool:
        """
        Test whether a point lies within the sphere.

        Parameters
        ----------
        p : quantity or array-like
            An array of shape (3,) (resp. (N, 3)) representing one (resp. N)
            points. If a unitless value is passed, it is interpreted as
            ``ucc['length']``.

        strict : bool
            If ``True``, comparison is done using strict inequalities (<, >).

        Returns
        -------
        result : array of bool or bool
            ``True`` iff ``p`` in within the sphere.
        """
        length_units = ucc.get("length")
        if self.to_world is not None:
            trafo = (
                self.to_world
                @ mi.Transform4f.translate(self.center.m_as(length_units))
                @ mi.Transform4f.scale(self.radius.m_as(length_units))
            )
        else:
            trafo = mi.Transform4f.translate(
                self.center.m_as(length_units)
            ) @ mi.Transform4f.scale(self.radius.m_as(length_units))
        p = np.atleast_2d(ensure_units(p, ucc.get("length")).m_as(length_units))
        c = trafo @ (0, 0, 0)
        d = np.linalg.norm(p - c, axis=1)
        r = np.linalg.norm(trafo @ (1, 0, 0) - c)
        return d < r if strict else d <= r

    @classmethod
    def surface(
        cls,
        altitude=0.0 * ureg.km,
        planet_radius: pint.Quantity = EARTH_RADIUS,
        bsdf: BSDF | None = None,
    ) -> SphereShape:
        """
        This class method constructor provides a simplified parametrization of
        the sphere shape better suited for the definition of the surface when
        configuring the one-dimensional model.

        The resulting sphere shape is centred at [0, 0, -`planet_radius`] and
        has a radius equal to `planet_radius` + `altitude`.

        Parameters
        ----------
        altitude : quantity or array-like, optional, default: 0 km
            Surface altitude. If a unitless value is passed, it is interpreted
            as ``ucc['length']``.

        planet_radius : quantity or float, optional
            Planet radius. If a unitless value is passed, it is interpreted
            as ``ucc['length']``. The default is Earth's radius.

        bsdf : BSDF or dict, optional, default: None
            A BSDF specification, forwarded to the main constructor.

        Returns
        -------
        SphereShape
            A sphere shape which can be used as the surface in a spherical shell
            geometry.
        """
        altitude = pinttr.util.ensure_units(altitude, default_units=ucc.get("length"))

        planet_radius = pinttr.util.ensure_units(
            planet_radius, default_units=ucc.get("length")
        )

        return cls(
            center=[0.0, 0.0, 0.0] * planet_radius.units,
            radius=planet_radius + altitude,
            bsdf=bsdf,
        )

    @classmethod
    def atmosphere(
        cls,
        top: pint.Quantity = 100.0 * ureg.km,
        planet_radius: pint.Quantity = EARTH_RADIUS,
        bsdf: BSDF | None = None,
    ) -> SphereShape:
        """
        This class method constructor provides a simplified parametrization of
        the sphere shape better suited for the definition of the surface when
        configuring the one-dimensional model.

        The resulting sphere shape is centred at [0, 0, 0] and
        has a radius equal to `planet_radius` + `top`.

        Parameters
        ----------
        top : quantity or array-like, optional, default: 100 km
            Top-of-atmosphere altitude. If a unitless value is passed, it is
            interpreted as ``ucc['length']``.

        planet_radius : quantity or float, optional
            Planet radius. If a unitless value is passed, it is interpreted
            as ``ucc['length']``. The default is Earth's radius.

        bsdf : BSDF or dict, optional, default: None
            A BSDF specification, forwarded to the main constructor.

        Returns
        -------
        SphereShape
            A sphere shape which can be used as the stencil of a participating
            medium in a spherical shell geometry.
        """
        top = pinttr.util.ensure_units(top, default_units=ucc.get("length"))

        planet_radius = pinttr.util.ensure_units(
            planet_radius, default_units=ucc.get("length")
        )

        return cls(
            center=[0.0, 0.0, 0.0] * planet_radius.units,
            radius=planet_radius + top,
            bsdf=bsdf,
        )



"""Exceptions and warnings specific to Eradiate."""

from pinttr.util import always_iterable

# ------------------------------------------------------------------------------
#                                   Exceptions
# ------------------------------------------------------------------------------


class ModeError(Exception):
    """Raised when encountering issues with Eradiate modes."""

    pass


class UnsetModeError(ModeError):
    """
    Raised when an operational mode is request but no active mode is selected.
    """

    def __str__(self):
        return (
            "No active mode selected. "
            "Select one by calling the `eradiate.set_mode()` function."
        )


class UnsupportedModeError(ModeError):
    """Raised when a feature is not supported for the currently selected mode."""

    def __init__(self, supported=None, unsupported=None, msg=None):
        super(UnsupportedModeError, self).__init__(msg)

        import eradiate

        self.mode = eradiate.mode().id if eradiate.mode() is not None else None
        self.supported = list(always_iterable(supported))
        self.unsupported = list(always_iterable(unsupported))

    def __str__(self):
        msg = f"'{self.mode}'" if self.mode is not None else "None"
        extra_msg = []

        if self.supported:
            extra_msg.append(f"supported: {', '.join(self.supported)}")

        if self.unsupported:
            extra_msg.append(f"unsupported: {', '.join(self.unsupported)}")

        if extra_msg:
            msg += f" ({'; '.join(extra_msg)})"

        return msg


class KernelVariantError(Exception):
    """Raised when encountering issues with Eradiate kernel variants."""

    pass


class DataError(Exception):
    """Raised when encountering issues with data."""

    pass


class TraversalError(Exception):
    """Raised when an error is encountered during scene element traversal."""

    pass


class InterpolationError(Exception):
    """Raised when encountering errors during interpolation."""

    pass


class MissingCoordinateError(InterpolationError):
    """Raised when a dataset is being interpolated along a coordinate that it
    does not have."""

    pass


class ScalarCoordinateError(InterpolationError):
    """Raised when a dataset is being interpolated along a scalar coordinate."""

    pass


class OutOfBoundsCoordinateError(InterpolationError):
    """Raised when a dataset is being interpolated along a coordinate with
    out-of-bounds values."""

    pass


# ------------------------------------------------------------------------------
#                                   Warnings
# ------------------------------------------------------------------------------


class ConfigWarning(UserWarning):
    """Used when encountering nonfatal configuration issues."""

    pass


class OverriddenValueWarning(UserWarning):
    """Used when a user-defined value is overridden during execution."""

    pass



from __future__ import annotations

from ._core import AbstractDirectionalIllumination
from ..core import NodeSceneElement
from ...attrs import define


@define(eq=False, slots=False)
class DirectionalIllumination(AbstractDirectionalIllumination):
    """
    Directional illumination scene element [``directional``].

    The illumination is oriented based on the classical angular convention used
    in Earth observation.
    """

    @property
    def template(self) -> dict:
        return {"type": "directional", "to_world": self._to_world}

    @property
    def objects(self) -> dict[str, NodeSceneElement]:
        return {"irradiance": self.irradiance}



"""
Particle layers.
"""

from __future__ import annotations

from functools import singledispatchmethod
from typing import Literal

import attrs
import numpy as np
import pint
import pinttr
import xarray as xr

from ._core import AbstractHeterogeneousAtmosphere
from ._particle_dist import ParticleDistribution, particle_distribution_factory
from ..core import traverse
from ..phase import TabulatedPhaseFunction
from ... import converters
from ...attrs import define, documented
from ...contexts import KernelContext
from ...kernel import SceneParameter
from ...radprops import ZGrid
from ...spectral.index import (
    CKDSpectralIndex,
    MonoSpectralIndex,
    SpectralIndex,
)
from ...units import to_quantity
from ...units import unit_context_config as ucc
from ...units import unit_registry as ureg
from ...util.misc import cache_by_id, summary_repr
from ...validators import is_positive


def _particle_layer_distribution_converter(value):
    if isinstance(value, str):
        if value == "uniform":
            return particle_distribution_factory.convert({"type": "uniform"})
        elif value == "gaussian":
            return particle_distribution_factory.convert({"type": "gaussian"})
        elif value == "exponential":
            return particle_distribution_factory.convert({"type": "exponential"})

    return particle_distribution_factory.convert(value)


@define(eq=False, slots=False)
class ParticleLayer(AbstractHeterogeneousAtmosphere):
    """
    Particle layer scene element [``particle_layer``].

    The particle layer has a vertical extension specified by a bottom altitude
    (set by ``bottom``) and a top altitude (set by ``top``).
    Inside the layer, the particles number is distributed according to a
    distribution (set by ``distribution``).
    See :mod:`~eradiate.scenes.atmosphere.particle_dist` for the available
    distribution types and corresponding parameters.
    The particle layer is itself divided into a number of (sub-)layers
    (``n_layers``) to allow to describe the variations of the particles number
    with altitude.
    The particle density in the layer is adjusted so that the particle layer's
    optical thickness at a specified reference wavelength (``w_ref``) meets a
    specified value (``tau_ref``).
    The particles radiative properties are specified by a data set
    (``dataset``).

    Notes
    -----
    If the optical property dataset contains only one spectral data point, the
    data is considered uniform throughout the entire spectrum.
    """

    bottom: pint.Quantity = documented(
        pinttr.field(
            default=ureg.Quantity(0.0, ureg.km),
            validator=[is_positive, pinttr.validators.has_compatible_units],
            units=ucc.deferred("length"),
        ),
        doc="Bottom altitude of the particle layer.\n"
        "\n"
        "Unit-enabled field (default: ucc[length])",
        type="quantity",
        init_type="float or quantity",
        default="0 km",
    )

    top: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("length"),
            default=ureg.Quantity(1.0, ureg.km),
            validator=[is_positive, pinttr.validators.has_compatible_units],
        ),
        doc="Top altitude of the particle layer.\n"
        "\n"
        "Unit-enabled field (default: ucc[length]).",
        type="quantity",
        init_type="float or quantity",
        default="1 km.",
    )

    @bottom.validator
    @top.validator
    def _bottom_top_validator(self, attribute, value):
        if self.bottom >= self.top:
            raise ValueError(
                f"while validating '{attribute.name}': bottom altitude must be "
                "lower than top altitude "
                f"(got bottom={self.bottom}, top={self.top})"
            )

    distribution: ParticleDistribution = documented(
        attrs.field(
            default="uniform",
            converter=_particle_layer_distribution_converter,
            validator=attrs.validators.instance_of(ParticleDistribution),
        ),
        doc="Particle distribution. Simple defaults can be set using a string: "
        '``"uniform"`` (resp. ``"gaussian"``, ``"exponential"``) is converted to '
        ":class:`UniformParticleDistribution() <.UniformParticleDistribution>` "
        "(resp. :class:`GaussianParticleDistribution() <.GaussianParticleDistribution>`, "
        ":class:`ExponentialParticleDistribution() <.ExponentialParticleDistribution>`).",
        init_type=":class:`.ParticleDistribution` or dict or "
        '{"uniform", "gaussian", "exponential"}, optional',
        type=":class:`.ParticleDistribution`",
        default='"uniform"',
    )

    w_ref: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("wavelength"),
            default=550 * ureg.nm,
            validator=[is_positive, pinttr.validators.has_compatible_units],
        ),
        doc="Reference wavelength at which the extinction optical thickness is "
        "specified.\n"
        "\n"
        "Unit-enabled field (default: ucc['wavelength']).",
        type="quantity",
        init_type="quantity or float",
        default="550.0",
    )

    tau_ref: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("dimensionless"),
            default=ureg.Quantity(0.2, ureg.dimensionless),
            validator=[is_positive, pinttr.validators.has_compatible_units],
        ),
        doc="Extinction optical thickness at the reference wavelength.\n"
        "\n"
        "Unit-enabled field (default: ucc[dimensionless]).",
        type="quantity",
        init_type="quantity or float",
        default="0.2",
    )

    dataset: xr.Dataset = documented(
        attrs.field(
            default="govaerts_2021-continental",
            converter=converters.passthrough_type(xr.Dataset)(
                attrs.converters.pipe(
                    converters.resolve_keyword(lambda x: f"aerosol/{x}.nc"),
                    converters.resolve_path,
                    converters.load_dataset,
                )
            ),
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Particle radiative property data set. "
        "If an xarray dataset is passed, the dataset is used as is "
        "(refer to the data guide for the format requirements of this dataset). "
        "If a path is passed, the converter looks it up on the hard drive, using "
        "the file resolver. "
        "If a string is passed, it is interpreted as a particle radiative "
        "property dataset identifier.",
        type="Dataset",
        init_type="Dataset or path-like or str",
        default='"govaerts_2021-continental"',
    )

    has_absorption: bool = documented(
        attrs.field(default=True, converter=bool),
        doc="Absorption bypass switch. If ``True``, the absorption coefficient "
        "is computed. Else, the absorption coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(default=True, converter=bool),
        doc="Scattering bypass switch. If ``True``, the scattering coefficient "
        "is computed. Else, the scattering coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    @has_absorption.validator
    @has_scattering.validator
    def _switch_validator(self, attribute, value):
        if not self.has_absorption and not self.has_scattering:
            raise ValueError(
                f"while validating {attribute.name}: at least one of "
                "'has_absorption' and 'has_scattering' must be True"
            )

    force_polarized_phase: bool = documented(
        attrs.field(default=False, converter=bool),
        doc="Force the use of a polarized phase function implementation, even"
        "when no polarization information is available.",
        type="bool",
        default="False",
    )

    particle_shape: Literal["spherical", "spheroidal"] = documented(
        attrs.field(default="spherical", kw_only=True),
        doc="Defines the shape of the particle. Only used in polarized mode.\n\n"
        '* ``"spherical"``: 4 coefficients considered [m11, m12, m33, m34].\n'
        '* ``"spheroidal"``: 6 coefficients considered [m11, m12, m22, m33, m34, m44].',
        type="str",
        init_type='{"spherical", "spheroidal"}',
        default='"spherical"',
    )

    _phase: TabulatedPhaseFunction | None = attrs.field(default=None, init=False)

    def update(self) -> None:
        self._phase = TabulatedPhaseFunction(
            id=self.phase_id,
            data=self.dataset.phase,
            force_polarized_phase=self.force_polarized_phase,
            particle_shape=self.particle_shape,
        )

    # --------------------------------------------------------------------------
    #                    Spatial and thermophysical properties
    # --------------------------------------------------------------------------

    def eval_fractions(self, zgrid: ZGrid) -> np.ndarray:
        """
        Compute the particle number fraction in the particle layer.

        Returns
        -------
        ndarray
            Particle number fractions as a (n_layers,)-shaped array.
        """
        x = (zgrid.layers - self.bottom) / (self.top - self.bottom)
        fractions = self.distribution(x.m_as(ureg.dimensionless))
        fractions /= np.sum(fractions)
        return fractions

    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        min_sigma_s = self.eval_sigma_s(ctx.si).min()
        return 1.0 / min_sigma_s if min_sigma_s != 0.0 else np.inf * ureg.m

    # --------------------------------------------------------------------------
    #                       Radiative properties
    # --------------------------------------------------------------------------

    @cache_by_id
    def _eval_albedo_impl(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Return albedo from dataset (without accounting for bypass switches)
        # This routine is vectorized and returns an array of shape
        # (n_wavelengths, n_layers)
        ds = self.dataset
        wavelengths = w.m_as(ds.w.attrs["units"])

        if len(ds["w"]) == 1:
            interpolated = to_quantity(ds.albedo.sel(w=wavelengths, method="nearest"))
        else:
            interpolated = to_quantity(ds.albedo.interp(w=wavelengths))
        where_present = np.reshape(self.eval_fractions(zgrid) > 0, (1, -1))
        return interpolated * where_present

    @cache_by_id
    def _eval_sigma_t_impl(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Return extinction coefficient from dataset (without accounting
        # for bypass switches). This routine is vectorized and returns an
        # array of shape (n_wavelengths, n_layers)

        # Collect input data
        ds = self.dataset
        ds_w_units = ureg(ds.w.attrs["units"])
        wavelengths = np.atleast_1d(w.m_as(ds_w_units))

        if len(ds["w"]) == 1:
            sigma_t_star = to_quantity(ds.sigma_t.sel(w=wavelengths, method="nearest"))
            sigma_t_star_ref = to_quantity(
                ds.sigma_t.sel(w=self.w_ref.m_as(ds_w_units), method="nearest")
            )
        else:
            sigma_t_star = to_quantity(ds.sigma_t.interp(w=wavelengths))
            sigma_t_star_ref = to_quantity(
                ds.sigma_t.interp(w=self.w_ref.m_as(ds_w_units))
            )

        # Compute target optical thickness value
        tau = self.tau_ref * sigma_t_star / sigma_t_star_ref

        # Scatter this total OT to all layers
        # TODO: Make sure that axis order is consistent with other vectorized
        #  routines
        fractions = self.eval_fractions(zgrid)
        tau_layers = np.broadcast_to(
            np.reshape(tau, (-1, 1)), (len(wavelengths), zgrid.n_layers)
        ) * np.reshape(fractions, (1, -1))

        # Compute corresponding average coefficient
        sigma_t = tau_layers / zgrid.layer_height

        return sigma_t

    def _eval_sigma_a_impl(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Return absorption coefficient from dataset (without accounting for
        # bypass switches). This routine is vectorized and returns an array of
        # shape (n_wavelengths, n_layers)
        albedo = self._eval_albedo_impl(w, zgrid)
        return self._eval_sigma_t_impl(w, zgrid) * (1.0 - albedo.m)

    def _eval_sigma_s_impl(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Return scattering coefficient from dataset (without accounting for
        # bypass switches). This routine is vectorized and returns an array of
        # shape (n_wavelengths, n_layers)
        albedo = self._eval_albedo_impl(w, zgrid)
        return self._eval_sigma_t_impl(w, zgrid) * albedo.m

    @singledispatchmethod
    def eval_albedo(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        raise NotImplementedError

    @eval_albedo.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_albedo_mono(
            w=si.w,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    @eval_albedo.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_albedo_ckd(
            w=si.w,
            g=si.g,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_albedo_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        if self.has_absorption and self.has_scattering:
            albedo = self._eval_albedo_impl(w, zgrid).squeeze()

        elif self.has_absorption and not self.has_scattering:
            albedo = 0.0 * ureg.dimensionless

        elif self.has_scattering and not self.has_absorption:
            albedo = 1.0 * ureg.dimensionless

        else:
            raise RuntimeError

        # Albedo is constant vs spatial dimension
        return np.full_like(zgrid.layers, albedo)

    def eval_albedo_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        return self.eval_albedo_mono(w=w, zgrid=zgrid)

    @singledispatchmethod
    def eval_sigma_t(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> pint.Quantity:
        # Inherit docstring
        raise NotImplementedError

    @eval_sigma_t.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_t_mono(
            w=si.w,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    @eval_sigma_t.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_t_ckd(
            w=si.w,
            g=si.g,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_t_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        result = self._eval_sigma_t_impl(w, zgrid).squeeze()

        if self.has_absorption and self.has_scattering:
            return result

        elif not self.has_absorption and self.has_scattering:
            return result - self._eval_sigma_a_impl(w, zgrid)

        elif self.has_absorption and not self.has_scattering:
            return result - self._eval_sigma_s_impl(w, zgrid)

        raise RuntimeError

    def eval_sigma_t_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        return self.eval_sigma_t_mono(w=w, zgrid=zgrid)

    @singledispatchmethod
    def eval_sigma_a(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> pint.Quantity:
        # Inherit docstring
        raise NotImplementedError

    @eval_sigma_a.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_a_mono(
            w=si.w,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    @eval_sigma_a.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_a_ckd(
            w=si.w,
            g=si.g,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_a_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        value = self._eval_sigma_a_impl(w, zgrid).squeeze()
        return value if self.has_absorption else np.zeros_like(value) * value.units

    def eval_sigma_a_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        return self.eval_sigma_a_mono(w, zgrid)

    @singledispatchmethod
    def eval_sigma_s(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        raise NotImplementedError

    @eval_sigma_s.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_s_mono(
            w=si.w,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    @eval_sigma_s.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_s_ckd(
            w=si.w,
            g=si.g,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_s_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        value = self._eval_sigma_s_impl(w, zgrid).squeeze()
        return value if self.has_scattering else np.zeros_like(value) * value.units

    def eval_sigma_s_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        return self.eval_sigma_s_mono(w, zgrid)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def phase(self) -> TabulatedPhaseFunction:
        # Inherit docstring
        return self._phase

    @property
    def _template_phase(self) -> dict:
        result, _ = traverse(self.phase)
        return result.data

    @property
    def _params_phase(self) -> dict[str, SceneParameter]:
        _, result = traverse(self.phase)
        return result.data



"""
Array radiative profile.
"""

from __future__ import annotations

import typing as t
import warnings

import attrs
import numpy as np
import pint
import xarray as xr

from ._core import RadProfile, ZGrid, make_dataset
from ..attrs import define, documented
from ..units import to_quantity
from ..units import unit_registry as ureg


@define(eq=False)
class ArrayRadProfile(RadProfile):
    """
    Array radiative profile.

    This class provides an interface to generate vertical profiles of
    atmospheric volume radiative properties (also sometimes referred to as
    collision coefficients).

    The array radiative profile is built from absorption and scattering
    coefficient data provided as input. This can be useful for debugging and
    benchmarking.
    """

    sigma_a: xr.DataArray | None = documented(
        attrs.field(
            default=None,
        ),
        doc="``DataArray`` of absorption coefficients. "
        "The ``DataArray`` is composed of two dimensions ``{w, z}`` representing the "
        "wavelength and altitude respectively. Note that ``w`` must be of length "
        "2 minimum to be correctly interpolated.",
        type="DataArray or None",
        init_type="DataArray or None",
        default="None",
    )

    sigma_s: xr.DataArray | None = documented(
        attrs.field(
            default=None,
        ),
        doc="``DataArray`` of scattering coefficients. "
        "The ``DataArray`` is composed of two dimensions ``{w, z}`` representing the "
        "wavelength and altitude respectively. Note that ``w`` must be of length "
        "2 minimum to be correctly interpolated.",
        type="DataArray or None",
        init_type="DataArray or None",
        default="None",
    )

    has_absorption: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Absorption switch. If ``True``, the absorption coefficient is "
        "computed. Else, the absorption coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Scattering switch. If ``True``, the scattering coefficient is "
        "computed. Else, the scattering coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    rayleigh_depolarization: np.ndarray = documented(
        attrs.field(
            converter=lambda x: np.array(x, dtype=np.float64),
            kw_only=True,
            factory=lambda: np.array(0.0),
        ),
        type="ndarray",
        doc="Depolarization factor of the rayleigh phase function. "
        "A ``ndarray`` will be interpreted as a description of the depolarization "
        "factor at different levels of the atmosphere. Must be shaped (N,) with "
        "N the number of layers.",
        init_type="array-like, optional",
        default="[0]",
    )

    interpolation_method: t.Literal["nearest", "linear"] = documented(
        attrs.field(
            default="nearest",
            converter=str,
            validator=attrs.validators.in_(["nearest", "linear"]),
        ),
        doc="Method of interpolation of the absorption and scattering coefficients.",
        type="str",
        init_type="{'nearest', 'linear'}",
        default="nearest",
    )

    interpolation_kwargs: dict[str, t.Any] = documented(
        attrs.field(
            factory=dict,
            converter=dict,
            validator=attrs.validators.instance_of(dict),
        ),
        doc="Keyword arguments passed to :meth:`xarray.DataArray.interp` when called.",
        type="dict",
        init_type="dict, optional",
    )

    def __attrs_post_init__(self):
        self.update()

    def update(self) -> None:
        pass

    @property
    def zbounds(self) -> tuple[pint.Quantity, pint.Quantity]:
        z_max = np.inf
        z_min = -1.0
        if self.sigma_s is not None and self.has_scattering:
            z_sigma_s = to_quantity(self.sigma_s.z)
            z_max = pint.Quantity(z_max, z_sigma_s.units)
            z_min = pint.Quantity(z_min, z_sigma_s.units)
            z_max = min(z_max, z_sigma_s.max())
            z_min = max(z_min, z_sigma_s.min())

        if self.sigma_a is not None and self.has_absorption:
            z_sigma_a = to_quantity(self.sigma_a.z)
            z_max = pint.Quantity(z_max, z_sigma_a.units)
            z_min = pint.Quantity(z_min, z_sigma_a.units)
            z_max = min(z_max, z_sigma_a.max())
            z_min = max(z_min, z_sigma_a.min())

        return z_min, z_max

    @property
    def levels(self) -> pint.Quantity:
        NotImplementedError

    @property
    def zgrid(self) -> ZGrid:
        NotImplementedError

    def eval_albedo_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Inherit docstring
        sigma_s = self.eval_sigma_s_mono(w, zgrid)
        sigma_t = self.eval_sigma_t_mono(w, zgrid)
        return np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        ).to("dimensionless")

    def eval_albedo_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        sigma_s = self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid)
        sigma_t = self.eval_sigma_t_ckd(w=w, g=g, zgrid=zgrid)
        return np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        ).to("dimensionless")

    def eval_sigma_a_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # NOTE: this method accepts 'w'-arrays and is vectorized as far as
        # each individual absorption dataset is concerned, namely when the
        # wavelengths span multiple datasets we for-loop over them.
        w = np.atleast_1d(w)
        if self.has_absorption and self.sigma_a is not None:
            values = self.sigma_a.interp(
                coords={
                    "z": zgrid.layers.m_as(self.sigma_a.z.attrs["units"]),
                    "w": w.m_as(self.sigma_a.w.attrs["units"]),
                },
                method=self.interpolation_method,
                kwargs=self.interpolation_kwargs,
            )

            values = to_quantity(values)
            return values.squeeze()
        else:
            return np.zeros((w.size, zgrid.n_layers)).squeeze() / ureg.km

    def eval_sigma_a_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        warnings.warn(
            "You are using ArrayRadProps in CKD mode, this is not standard "
            "behaviour and might provide erroneous results."
        )
        return self.eval_sigma_a_mono(w=w, zgrid=zgrid)

    def eval_sigma_s_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        w = np.atleast_1d(w)
        if self.has_scattering and self.sigma_s is not None:
            sigma_s = self.sigma_s.interp(
                coords={
                    "z": zgrid.layers.m_as(self.sigma_s.z.attrs["units"]),
                    "w": w.m_as(self.sigma_s.w.attrs["units"]),
                },
                method=self.interpolation_method,
                kwargs=self.interpolation_kwargs,
            )

            sigma_s = to_quantity(sigma_s)
            return sigma_s.squeeze()
        else:
            return np.zeros((1, zgrid.n_layers)) / ureg.km

    def eval_sigma_s_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        return self.eval_sigma_s_mono(w=w, zgrid=zgrid)

    def eval_sigma_t_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        sigma_a = self.eval_sigma_a_mono(w=w, zgrid=zgrid)
        sigma_s = self.eval_sigma_s_mono(w=w, zgrid=zgrid)
        return sigma_a + sigma_s

    def eval_sigma_t_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        sigma_a = self.eval_sigma_a_ckd(w=w, g=g, zgrid=zgrid)
        sigma_s = self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid)
        return sigma_a + sigma_s

    def eval_dataset_mono(self, w: pint.Quantity, zgrid: ZGrid) -> xr.Dataset:
        return make_dataset(
            wavelength=w,
            z_level=zgrid.levels,
            z_layer=zgrid.layers,
            sigma_a=self.eval_sigma_a_mono(w, zgrid),
            sigma_s=self.eval_sigma_s_mono(w, zgrid),
        ).squeeze()

    def eval_dataset_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> xr.Dataset:
        return make_dataset(
            wavelength=w,
            z_level=zgrid.levels,
            z_layer=zgrid.layers,
            sigma_a=self.eval_sigma_a_ckd(w=w, g=g, zgrid=zgrid),
            sigma_s=self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid),
        ).squeeze()

    def eval_depolarization_factor_mono(
        self, w: pint.Quantity, zgrid: ZGrid
    ) -> pint.Quantity:
        if self.has_scattering:
            if isinstance(self.rayleigh_depolarization, np.ndarray):
                return np.atleast_1d(self.rayleigh_depolarization) * ureg.dimensionless
            else:
                raise NotImplementedError

        else:
            return np.atleast_1d(0.0) * ureg.dimensionless

    def eval_depolarization_factor_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        return self.eval_depolarization_factor_mono(w=w, zgrid=zgrid)



import numpy as np

from eradiate import unit_registry as ureg
from eradiate.experiments import AtmosphereExperiment


def create_rami4atm_hom00_bla_sd2s_m03_z30a000_brfpp():
    r"""
    *RAMI4ATM HOM00_BLA_SD2S_M03*

    This scenario is based on the ``HOM00_BLA_SD2S_M03_z30a000-brfpp`` scenario
    of the RAMI4ATM benchmark.

    *Scene setup*

    - Geometry: 1D plane-parallel
    - Atmosphere: Molecular atmosphere using the AFGL 1986 (U.S. Standard) profile
    - Aerosol layer: Uniform layer ranging from 0 km to 2 km, with AOT at 550 nm = 0.5; aerosol dataset ``govaerts_2021-desert``
    - Surface: Black
    - Illumination: Directional illumination with a zenith angle of 30°
    - Sensor: Multi-distant measure covering the principal plane, from -75° to 75° with 2° increments, delta SRF positioned at λ = 550 nm
    """
    config = {
        "surface": {
            "reflectance": {"value": 0.0, "type": "uniform"},
            "type": "lambertian",
        },
        "atmosphere": {
            "molecular_atmosphere": {
                "has_absorption": False,
                "has_scattering": True,
                "type": "molecular",
                "thermoprops": {
                    "identifier": "afgl_1986-us_standard",
                    "z": np.arange(0, 120.05, 0.05) * ureg.km,
                },
                "absorption_data": "monotropa",
            },
            "particle_layers": [
                {
                    "bottom": 0,
                    "bottom_units": "meter",
                    "top": 2000,
                    "top_units": "meter",
                    "distribution": {"type": "uniform"},
                    "tau_ref": 0.2,
                    "dataset": "govaerts_2021-desert",
                }
            ],
            "type": "heterogeneous",
        },
        "illumination": {
            "zenith": 30.0,
            "zenith_units": "degree",
            "azimuth": 0.0,
            "azimuth_units": "degree",
            "type": "directional",
        },
        "measures": [
            {
                "type": "mdistant",
                "construct": "hplane",
                "zeniths": np.arange(-75, 76, 2),
                "zeniths_units": "degree",
                "azimuth": 0.0,
                "azimuth_units": "degree",
                "srf": "sentinel_2a-msi-3",
                "spp": 1000,
            }
        ],
        "ckd_quad_config": {
            "policy": "fixed",
            "type": "gauss_legendre",
            "ng_max": 16,
        },
        "integrator": {"type": "piecewise_volpath", "moment": True},
    }

    return AtmosphereExperiment(**config)



from __future__ import annotations

import logging
import typing as t
import warnings

import attrs
import drjit as dr
import mitsuba as mi
from mitsuba.python.util import SceneParameters as _MitsubaSceneParameters
from tqdm.auto import tqdm

from ._kernel_dict import KernelSceneParameterMap
from .. import config
from ..attrs import define, documented, frozen
from ..contexts import KernelContext
from ..rng import SeedState, root_seed_state

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
#                         Parameter lookup strategies
# ------------------------------------------------------------------------------


@frozen
class SearchSceneParameter:
    """
    This class defines a scene parameter search protocol applied during the
    Mitsuba scene tree traversal.

    Instances are callables which take, as argument, the current node during
    a Mitsuba scene tree traversal and, optionally, its path in the Mitsuba
    scene tree. If the lookup succeeds, the full parameter path is returned.

    See Also
    --------
    :func:`.mi_traverse`
    """

    node_type: type = documented(
        attrs.field(validator=attrs.validators.instance_of(type)),
        doc="Type of the node which will be looked up.",
        type="type",
    )

    node_id: str = documented(
        attrs.field(validator=attrs.validators.instance_of(str)),
        doc="ID of the node which will be looked up.",
        type="str",
    )

    parameter_relpath: str = documented(
        attrs.field(validator=attrs.validators.instance_of(str)),
        doc="Parameter path relative to its parent object.",
        type="str",
    )

    def __call__(self, node, node_path: str | None = None) -> str | None:
        if isinstance(node, self.node_type) and node.id() == self.node_id:
            prefix = f"{node_path}." if node_path is not None else ""
            return f"{prefix}{self.parameter_relpath}"

        else:
            return None


# ------------------------------------------------------------------------------
#                           Mitsuba scene traversal
# ------------------------------------------------------------------------------


@define
class MitsubaObjectWrapper:
    """
    This container aggregates a Mitsuba object, its associated parameters and a
    set of updaters that can be used to modify the scene parameters.

    See Also
    --------
    :func:`mi_traverse`
    """

    obj: mi.Object = documented(
        attrs.field(repr=lambda x: "Scene[...]" if isinstance(x, mi.Scene) else str(x)),
        doc="Mitsuba object.",
        type="mitsuba.Object",
    )

    parameters: mi.SceneParameters | None = documented(
        attrs.field(
            default=None,
            repr=lambda x: (
                "SceneParameters[...]"
                if x.__class__.__name__ == "SceneParameters"
                else str(None)
            ),
        ),
        doc="Mitsuba scene parameter map.",
        type="mitsuba.SceneParameters",
        init_type="mitsuba.SceneParameters, optional",
        default="None",
    )

    umap_template: KernelSceneParameterMap | None = documented(
        attrs.field(
            default=None,
            repr=lambda x: (
                "KernelSceneParameterMap[...]"
                if isinstance(x, KernelSceneParameterMap)
                else str(x)
            ),
        ),
        doc="An update map template, which can be rendered and used to update "
        "Mitsuba scene parameters depending on context information.",
        type=".KernelSceneParameterMap",
        init_type=".KernelSceneParameterMap, optional",
        default="None",
    )

    def drop_parameters(self) -> None:
        """
        Reduce the size of the scene parameter table :attr:`.parameters` by
        only keeping elements whose keys are listed in the parameter update
        map template :attr:`.umap_template`. For parameters associated with a
        lookup protocol, the looked up parameter ID is checked and used.
        """
        if self.umap_template is not None:
            keys = []
            for name, param in self.umap_template.items():
                if param.search is not None:
                    if param.parameter_id is not None:
                        keys.append(param.parameter_id)
                    else:
                        warnings.warn(
                            f"Parameter '{name}' has a lookup strategy but the "
                            "associated parameter ID is undefined; was a "
                            "parameter lookup performed during the Mitsuba "
                            "scene traversal?"
                        )
                else:
                    keys.append(name)

            self.parameters.keep(keys)


class SceneParameters(_MitsubaSceneParameters):
    def __init__(self, properties=None, hierarchy=None, aliases=None):
        super().__init__(properties, hierarchy)
        self.aliases = aliases if aliases is not None else {}

    def set_dirty(self, key: str):
        # Inherit docstring

        value, _, node, flags = self.properties[key]

        is_nondifferentiable = flags & mi.ParamFlags.NonDifferentiable.value
        if is_nondifferentiable and dr.grad_enabled(value):
            mi.Log(
                mi.LogLevel.Warn,
                f"Parameter '{key}' is marked as non-differentiable but has "
                "gradients enabled, unexpected results may occur!",
            )

        node_key = key  # Key of current node
        while node is not None:
            parent, depth = self.hierarchy[node]

            name = node_key
            if parent is not None:
                if "." not in name and depth > 0:
                    # We've hit the top level from an ID-aliased node:
                    # Resolve the alias to finish climbing the hierarchy
                    node_key = self.aliases[name]
                node_key, name = node_key.rsplit(".", 1)

            self.nodes_to_update.setdefault((depth, node), set())
            self.nodes_to_update[(depth, node)].add(name)

            node = parent

        return self.properties[key]


def mi_traverse(
    obj: mi.Object,
    umap_template: KernelSceneParameterMap | None = None,
    name_id_override: str | list[str] | bool | None = None,
) -> MitsubaObjectWrapper:
    """
    Traverse a node of the Mitsuba scene graph and return a container holding
    the Mitsuba scene, its parameter map and an updated parameter update map.

    Parameters
    ----------
    obj : mitsuba.Object
        Mitsuba scene graph node to be traversed.

    umap_template : .KernelSceneParameterMap, optional
        An additional update map template which is to be updated during
        traversal. This is used to perform parameter lookup during traversal.

    name_id_override : str or list of str, optional
        If set, this argument will be used to select nodes in the scene tree
        whose names will be "pinned" to their ID. Passed values are used as
        regular expressions, with all that it implies regarding ID string
        matching. If this parameter is set to ``True``, a regex that matches
        anything is used.

    Returns
    -------
    MitsubaObjectWrapper
        A container holding the traversed object, the corresponding parameter
        map and the parameter update map (if any).

    Notes
    -----
    This is a reimplementation of the :func:`mitsuba.traverse` function.
    """

    umap_template = (
        KernelSceneParameterMap(data=umap_template.data.copy())
        if umap_template is not None
        else KernelSceneParameterMap()
    )

    lookups = {
        k: v
        for k, v in umap_template.items()
        if v.parameter_id is None and v.search is not None
    }

    if name_id_override is None or name_id_override is False:
        name_id_override = []

    if name_id_override is True:
        name_id_override = [r".*"]

    if type(name_id_override) is not list:
        name_id_override = [name_id_override]

    import re

    regexps = [re.compile(k).match for k in name_id_override]

    class SceneTraversal(mi.TraversalCallback):
        def __init__(
            self,
            node,
            parent=None,
            properties=None,
            hierarchy=None,
            prefixes=None,
            name=None,
            depth=0,
            flags=+mi.ParamFlags.Differentiable,
            aliases=None,
        ):
            mi.TraversalCallback.__init__(self)
            self.properties = dict() if properties is None else properties
            self.hierarchy = dict() if hierarchy is None else hierarchy
            self.prefixes = set() if prefixes is None else prefixes
            self.aliases = dict() if aliases is None else aliases

            node_id = node.id()
            if name_id_override and node_id:
                for r in regexps:
                    if r(node_id):
                        if node_id != name:
                            self.aliases[node_id] = name
                        name = node_id
                        break

            if name is not None:
                ctr, name_len = 1, len(name)
                while name in self.prefixes:
                    name = f"{name[:name_len]}_{ctr}"
                    ctr += 1
                self.prefixes.add(name)

            self.name = name
            self.node = node
            self.depth = depth
            self.hierarchy[node] = (parent, depth)
            self.flags = flags

            # Try and recover a parameter ID from this node
            for name, uparam in list(lookups.items()):
                lookup_result = uparam.search(self.node, self.name)
                if lookup_result is not None:
                    uparam.parameter_id = lookup_result
                    del lookups[
                        name
                    ]  # Remove successful lookups to accelerate future searches

        def put_parameter(self, name, ptr, flags, cpptype=None):
            name = name if self.name is None else self.name + "." + name

            flags = self.flags | flags
            # Non-differentiable parameters shouldn't be flagged as discontinuous
            if (flags & mi.ParamFlags.NonDifferentiable) != 0:
                flags = flags & ~mi.ParamFlags.Discontinuous

            self.properties[name] = (ptr, cpptype, self.node, self.flags | flags)

        def put_object(self, name, node, flags):
            if node is None or node in self.hierarchy:
                return
            cb = SceneTraversal(
                node=node,
                parent=self.node,
                properties=self.properties,
                hierarchy=self.hierarchy,
                prefixes=self.prefixes,
                name=name if self.name is None else f"{self.name}.{name}",
                depth=self.depth + 1,
                flags=self.flags | flags,
                aliases=self.aliases,
            )
            node.traverse(cb)

    cb = SceneTraversal(obj)
    obj.traverse(cb)

    # Check if there are unsuccessful lookups
    if lookups:
        warnings.warn(
            "There were unsuccessful Mitsuba scene parameter lookups: "
            f"{list(lookups.keys())}"
        )

    return MitsubaObjectWrapper(
        obj=obj,
        parameters=SceneParameters(cb.properties, cb.hierarchy, cb.aliases),
        umap_template=umap_template,
    )


# ------------------------------------------------------------------------------
#                             Mitsuba scene render
# ------------------------------------------------------------------------------


def mi_render(
    mi_scene: MitsubaObjectWrapper,
    ctxs: list[KernelContext],
    sensors: None | int | list[int] = None,
    spp: int = 0,
    seed_state: SeedState | None = None,
) -> dict[t.Any, mi.Bitmap]:
    """
    Render a Mitsuba scene multiple times given specified contexts and sensor
    indices.

    Parameters
    ----------
    mi_scene : .MitsubaObjectWrapper
        Mitsuba scene to render.

    ctxs : list of :class:`.KernelContext`
        List of contexts used to generate the parameter update table at each
        iteration.

    sensors : int or list of int, optional
        Sensor indices to render. If ``None`` (default), all sensors are
        rendered.

    spp : int, optional, default: 0
        Number of samples per pixel. If set to 0 (default), the value set in the
        original scene definition takes precedence.

    seed_state : .SeedState, optional
        Seed state used to generate seeds to initialize Mitsuba's RNG at
        each run. If unset, Eradiate's root seed state is used.

    Returns
    -------
    dict
        A nested dictionary mapping context and sensor indices to rendered
        bitmaps.

    Notes
    -----
    This function wraps sequential calls to  :func:`mitsuba.render`.
    """

    if seed_state is None:
        logger.debug("Using default RNG seed generator")
        seed_state = root_seed_state

    results = {}

    # Loop on contexts
    with tqdm(
        initial=0,
        total=len(ctxs),
        unit_scale=1.0,
        leave=True,
        bar_format="{desc}{n:g}/{total:g}|{bar}| {elapsed}, ETA={remaining}",
        disable=(config.settings.progress < config.ProgressLevel.SPECTRAL_LOOP)
        or len(ctxs) <= 1,
    ) as pbar:
        for ctx in ctxs:
            pbar.set_description(
                f"Eradiate [{ctx.index_formatted}]",
                refresh=True,
            )

            logger.debug("Updating Mitsuba scene parameters")
            mi_scene.parameters.update(mi_scene.umap_template.render(ctx))

            if sensors is None:
                mi_sensors = [
                    (i, sensor) for i, sensor in enumerate(mi_scene.obj.sensors())
                ]

            else:
                if isinstance(sensors, int):
                    sensors = [sensors]
                mi_sensors = [(i, mi_scene.obj.sensors()[i]) for i in sensors]

            # Loop on sensors
            for i_sensor, mi_sensor in mi_sensors:
                # Render sensor
                seed = int(seed_state.next().squeeze())
                logger.debug(
                    'Running Mitsuba for sensor "%s" with seed value %s',
                    mi_sensor.id(),
                    seed,
                )
                mi.render(mi_scene.obj, sensor=i_sensor, seed=seed, spp=spp)

                # Store result in a new Bitmap object
                siah = ctx.si.as_hashable
                if siah not in results:
                    results[siah] = {}

                results[siah][mi_sensor.id()] = mi.Bitmap(mi_sensor.film().bitmap())

            pbar.update()

    return results
